from typing import Annotated, Literal
from httpx import Client
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.exceptions import NonRetryableActionError
from admyral.secret.secret import register_secret


@register_secret(secret_type="Zendesk")
class ZendeskSecret(BaseModel):
    subdomain: str
    email: str
    api_token: str


def get_zendesk_client(secret: ZendeskSecret) -> Client:
    return Client(
        base_url=f"https://{secret.subdomain}.zendesk.com/api",
        headers={
            "Content-Type": "application/json",
        },
        auth=(f"{secret.email}/token", secret.api_token),
    )


@action(
    display_name="List Users",
    display_namespace="Zendesk",
    description="List users from Zendesk ticketing system.",
    secrets_placeholders=["ZENDESK_SECRET"],
)
def list_zendesk_users(
    user_status: Annotated[
        Literal["active", "inactive"] | None,
        ArgumentMetadata(
            display_name="User Status",
            description="The status of the users to list. Possible values: active, inactive",
        ),
    ] = None,
    user_role: Annotated[
        list[Literal["end-user", "agent", "admin"]]
        | Literal["end-user", "agent", "admin"]
        | None,
        ArgumentMetadata(
            display_name="User Role",
            description="Role of user to search for. Possible values: user, admin, end-user. You can "
            'either filter for one role (e.g., admin) or for multiple roles (e.g., ["admin", "end-user"]).',
        ),
    ] = None,
) -> JsonValue:
    # https://developer.zendesk.com/api-reference/ticketing/users/users/
    secret = ctx.get().secrets.get("ZENDESK_SECRET")
    secret = ZendeskSecret.model_validate(secret)

    if user_role is not None:
        if isinstance(user_role, str) and user_role not in [
            "end-user",
            "agent",
            "admin",
        ]:
            raise NonRetryableActionError(
                f"Invalid user role: {user_role}. If user role is defined, it must be one of: end-user, agent, admin"
            )
        else:
            if not isinstance(user_role, list):
                raise NonRetryableActionError(
                    "User role must be a list of roles or a single role defined as a string."
                )
            for role in user_role:
                if role not in ["end-user", "agent", "admin"]:
                    raise NonRetryableActionError(
                        f"Invalid user role: {role}. If user role is defined, it must be one of: end-user, agent, admin"
                    )

    with get_zendesk_client(secret) as client:
        page = 1
        endpoint = "/v2/users.json?per_page=100&page={page}"

        if user_role is not None:
            if isinstance(user_role, list):
                roles = "&".join(f"role[]={user_role}")
            else:
                roles = f"role={user_role}"
            endpoint += f"&{roles}"

        users = []

        while True:
            response = client.get(endpoint.format(page=page))
            response.raise_for_status()
            data = response.json()
            users_page = data["users"]
            if user_status is None:
                users.extend(users_page)
            else:
                users.extend(
                    [
                        user
                        for user in users_page
                        if (user_status == "active" and user["verified"])
                        or (user_status == "inactive" and not user["verified"])
                    ]
                )

            if len(users_page) < 100:
                break
            page += 1

        return users



"""add canceled_at to WorkflowRun schema

Revision ID: 5b9dd35eb10b
Revises: 5ebc9fcfc1c7
Create Date: 2024-09-05 21:16:01.841467

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "5b9dd35eb10b"
down_revision: Union[str, None] = "5ebc9fcfc1c7"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "workflow_runs", sa.Column("canceled_at", sa.TIMESTAMP(), nullable=True)
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("workflow_runs", "canceled_at")
    # ### end Alembic commands ###



"""add auth schema

Revision ID: 552e7a8dec07
Revises: 5b9dd35eb10b
Create Date: 2024-10-02 14:48:49.354845

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel  # noqa F401


# revision identifiers, used by Alembic.
revision: str = "552e7a8dec07"
down_revision: Union[str, None] = "5b9dd35eb10b"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "User",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("id", sa.TEXT(), nullable=False),
        sa.Column("name", sa.TEXT(), nullable=True),
        sa.Column("email", sa.TEXT(), nullable=False),
        sa.Column("email_verified", sa.TIMESTAMP(), nullable=True),
        sa.Column("image", sa.TEXT(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("email"),
    )
    op.create_table(
        "VerificationToken",
        sa.Column("identifier", sa.TEXT(), nullable=False),
        sa.Column("token", sa.TEXT(), nullable=False),
        sa.Column("expires", sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint("identifier", "token"),
        sa.UniqueConstraint("identifier", "token"),
    )
    op.create_table(
        "Account",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("userId", sa.TEXT(), nullable=False),
        sa.Column("type", sa.TEXT(), nullable=False),
        sa.Column("provider", sa.TEXT(), nullable=False),
        sa.Column("providerAccountId", sa.TEXT(), nullable=False),
        sa.Column("refresh_token", sa.TEXT(), nullable=True),
        sa.Column("access_token", sa.TEXT(), nullable=True),
        sa.Column("expires_at", sa.Integer(), nullable=True),
        sa.Column("token_type", sa.TEXT(), nullable=True),
        sa.Column("scope", sa.TEXT(), nullable=True),
        sa.Column("id_token", sa.TEXT(), nullable=True),
        sa.Column("session_state", sa.TEXT(), nullable=True),
        sa.ForeignKeyConstraint(["userId"], ["User.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("provider", "providerAccountId"),
        sa.UniqueConstraint("provider", "providerAccountId"),
    )
    op.create_table(
        "Authenticator",
        sa.Column("credentialID", sa.TEXT(), nullable=False),
        sa.Column("userId", sa.TEXT(), nullable=False),
        sa.Column("providerAccountId", sa.TEXT(), nullable=False),
        sa.Column("credentialPublicKey", sa.TEXT(), nullable=False),
        sa.Column("counter", sa.Integer(), nullable=False),
        sa.Column("credentialDeviceType", sa.TEXT(), nullable=False),
        sa.Column("credentialBackedUp", sa.BOOLEAN(), nullable=False),
        sa.Column("transports", sa.TEXT(), nullable=True),
        sa.ForeignKeyConstraint(["userId"], ["User.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("credentialID", "userId"),
        sa.UniqueConstraint("credentialID"),
    )
    op.create_table(
        "Session",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("id", sa.TEXT(), nullable=False),
        sa.Column("sessionToken", sa.TEXT(), nullable=False),
        sa.Column("userId", sa.TEXT(), nullable=False),
        sa.Column("expires", sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(["userId"], ["User.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("sessionToken"),
    )
    op.create_table(
        "api_keys",
        sa.Column(
            "created_at",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column(
            "updated_at",
            sa.TIMESTAMP(),
            server_default=sa.text("now()"),
            nullable=False,
        ),
        sa.Column("id", sa.TEXT(), nullable=False),
        sa.Column("user_id", sa.TEXT(), nullable=False),
        sa.Column("name", sa.TEXT(), nullable=False),
        sa.Column("key", sa.TEXT(), nullable=False),
        sa.ForeignKeyConstraint(["user_id"], ["User.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_api_keys_key"), "api_keys", ["key"], unique=False)
    op.create_index(op.f("ix_api_keys_user_id"), "api_keys", ["user_id"], unique=False)

    DEFAULT_USER_ID = "38815447-e272-4299-94c0-29a2d30435f9"
    op.execute(
        f"INSERT INTO \"User\" (id, email) VALUES ('{DEFAULT_USER_ID}', 'random@email.com')",
    )
    TEST_USER_ID = "a2f038f1-e35b-4509-bcc4-c08bd0e481a6"
    op.execute(
        f"INSERT INTO \"User\" (id, email) VALUES ('{TEST_USER_ID}', 'test@admyral.ai')",
    )

    op.execute(f"UPDATE secrets SET user_id = '{DEFAULT_USER_ID}'")
    op.create_foreign_key(
        None, "secrets", "User", ["user_id"], ["id"], ondelete="CASCADE"
    )
    op.drop_column("workflow_runs", "user_id")

    op.execute(f"UPDATE workflows SET user_id = '{DEFAULT_USER_ID}'")
    op.create_foreign_key(
        None, "workflows", "User", ["user_id"], ["id"], ondelete="CASCADE"
    )

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "workflows", type_="foreignkey")
    op.add_column(
        "workflow_runs",
        sa.Column("user_id", sa.TEXT(), autoincrement=False, nullable=False),
    )
    op.drop_constraint(None, "secrets", type_="foreignkey")
    op.drop_index(op.f("ix_api_keys_user_id"), table_name="api_keys")
    op.drop_index(op.f("ix_api_keys_key"), table_name="api_keys")
    op.drop_table("api_keys")
    op.drop_table("Session")
    op.drop_table("Authenticator")
    op.drop_table("Account")
    op.drop_table("VerificationToken")
    op.drop_table("User")
    # ### end Alembic commands ###



from typing import Annotated, Literal
import base64
from httpx import Client
from pydantic import BaseModel

from admyral.typings import JsonValue
from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.utils.collections import is_empty
from admyral.secret.secret import register_secret
from admyral.exceptions import NonRetryableActionError


@register_secret(secret_type="Jira")
class JiraSecret(BaseModel):
    domain: str
    email: str
    api_key: str


def get_jira_client(secret: JiraSecret) -> Client:
    api_key_base64 = base64.b64encode(
        f"{secret.email}:{secret.api_key}".encode()
    ).decode()
    return Client(
        base_url=f"https://{secret.domain}/rest/api/3",
        headers={
            "Authorization": f"Basic {api_key_base64}",
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Create Jira Issue",
    display_namespace="Jira",
    description="Create a Jira issue",
    secrets_placeholders=["JIRA_SECRET"],
)
def create_jira_issue(
    summary: Annotated[
        str,
        ArgumentMetadata(
            display_name="Summary",
            description="The summary of the issue",
        ),
    ],
    project_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="Project ID",
            description="The ID of the project to create the issue in",
        ),
    ],
    issue_type: Annotated[
        str,
        ArgumentMetadata(
            display_name="Issue Type",
            description="The issue type",
        ),
    ],
    description: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Description",
            description="The description of the issue in Atlassian Document Format",
        ),
    ] = None,
    assignee: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Assignee",
            description="The ID of the assignee",
        ),
    ] = None,
    labels: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Labels",
            description="The labels to add to the issue",
        ),
    ] = None,
    priority: Annotated[
        Literal["Lowest", "Low", "Medium", "High", "Highest"] | None,
        ArgumentMetadata(
            display_name="Priority",
            description="The priority of the issue",
        ),
    ] = None,
    custom_fields: Annotated[
        dict[str, JsonValue] | None,
        ArgumentMetadata(
            display_name="Custom Fields",
            description="The custom fields to add to the issue as a JSON object",
        ),
    ] = None,
) -> JsonValue:
    """
    How to find the project id?

        curl --request GET \
            --url https://christesting123.atlassian.net/rest/api/latest/project  \
            --header 'Authorization: Basic <base64-encoded-email:api_key-pair>'


    How to find the user id?

        curl --request GET \
            --url https://<your_domain>.atlassian.net/rest/api/3/myself  \
            --header 'Authorization: Basic <base64-encoded-email:api_key-pair>'
        
        or

        go to https://<your_domain>.atlassian.net/people
    
    How to find the issue type id?

        curl --request GET \
            --url https://<your_domain>.atlassian.net/rest/api/3/issuetype  \
            --header 'Authorization: Basic <base64-encoded-email:api_key-pair>'

    """
    # Atlassian Document Format: https://developer.atlassian.com/cloud/jira/platform/apis/document/structure/
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issues/#api-rest-api-3-issue-post
    secret = ctx.get().secrets.get("JIRA_SECRET")
    secret = JiraSecret.model_validate(secret)

    body = {
        "fields": {
            "project": {
                "id": project_id,
            },
            "summary": summary,
            "issuetype": {
                "name": issue_type,
            },
        }
    }

    if description:
        body["fields"]["description"] = description

    if assignee:
        body["fields"]["assignee"] = {
            "id": assignee,
        }

    if labels:
        body["fields"]["labels"] = labels

    if priority:
        body["fields"]["priority"] = {
            "name": priority,
        }

    if custom_fields:
        for key, value in custom_fields.items():
            body["fields"][key] = value

    with get_jira_client(secret) as client:
        response = client.post(
            "/issue",
            json=body,
        )
        response.raise_for_status()
        return response.json()


@action(
    display_name="Update Jira Issue Status",
    display_namespace="Jira",
    description="Update a Jira issue status",
    secrets_placeholders=["JIRA_SECRET"],
)
def update_jira_issue_status(
    issue_id_or_key: Annotated[
        str,
        ArgumentMetadata(
            display_name="Issue ID or Key",
            description="The ID or the key of the issue",
        ),
    ],
    status: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Status",
            description="The status to set the issue to",
        ),
    ],
) -> None:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issues/#api-rest-api-3-issue-issueidorkey-transitions-post
    # https://community.atlassian.com/t5/Jira-questions/How-do-i-find-a-transition-ID/qaq-p/2113213
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issues/#api-rest-api-3-issue-issueidorkey-transitions-get
    secret = ctx.get().secrets.get("JIRA_SECRET")
    secret = JiraSecret.model_validate(secret)

    with get_jira_client(secret) as client:
        response = client.get(f"/issue/{issue_id_or_key}/transitions")
        response.raise_for_status()
        transitions = response.json()
        transition_id = next(
            (
                t["id"]
                for t in transitions.get("transitions", [])
                if t["name"].lower() == status.lower()
            ),
            None,
        )
        if transition_id is None:
            raise NonRetryableActionError(
                f"Could not move the Jira issue into status {status} because there is no transition "
                "into this status from the current status."
            )

        response = client.post(
            f"/issue/{issue_id_or_key}/transitions",
            json={
                "transition": {
                    "id": transition_id,
                }
            },
        )
        response.raise_for_status()


@action(
    display_name="Comment Jira Issue",
    display_namespace="Jira",
    description="Comment Jira issue",
    secrets_placeholders=["JIRA_SECRET"],
)
def comment_jira_issue_status(
    issue_id_or_key: Annotated[
        str,
        ArgumentMetadata(
            display_name="Issue ID or Key",
            description="The ID or the key of the issue",
        ),
    ],
    comment: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Comment",
            description="The comment in Atlassian Document Format",
        ),
    ] = None,
) -> None:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issues/#api-rest-api-3-issue-issueidorkey-transitions-post
    secret = ctx.get().secrets.get("JIRA_SECRET")
    secret = JiraSecret.model_validate(secret)

    with get_jira_client(secret) as client:
        response = client.post(
            f"/issue/{issue_id_or_key}/comment",
            json={"body": comment},
        )
        response.raise_for_status()


@action(
    display_name="Jira Issue Search",
    display_namespace="Jira",
    description="Search Jira issues using JQL",
    secrets_placeholders=["JIRA_SECRET"],
)
def search_jira_issues(
    jql: Annotated[
        str,
        ArgumentMetadata(
            display_name="JQL",
            description="The JQL query to search for issues",
        ),
    ],
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of issues to return",
        ),
    ] = 1000,
) -> list[JsonValue]:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issue-search/#api-rest-api-3-search-get
    secret = ctx.get().secrets.get("JIRA_SECRET")
    secret = JiraSecret.model_validate(secret)

    with get_jira_client(secret) as client:
        offset = 0
        issues = []

        while limit is None or len(issues) < limit:
            response = client.get(
                "/search",
                params={"jql": jql, "maxResults": 100, "startAt": offset},
            )
            response.raise_for_status()
            data = response.json()

            issues.extend(data["issues"])
            offset = len(issues)
            if offset == data["total"]:
                break

        return issues if limit is None else issues[:limit]


@action(
    display_name="Get Audit Records",
    display_namespace="Jira",
    description="Get audit records",
    secrets_placeholders=["JIRA_SECRET"],
)
def get_jira_audit_records(
    start_date: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Date",
            description="The start date in in ISO-8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_date: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Date",
            description="The end date in in ISO-8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    filter: Annotated[
        list[str] | None,
        ArgumentMetadata(
            display_name="Filter",
            description="A list of strings to match with audit field content. The strings must not contain spaces.",
        ),
    ] = None,
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of audit records to return.",
        ),
    ] = None,
):
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-audit-records/#api-rest-api-3-auditing-record-get
    secret = ctx.get().secrets.get("JIRA_SECRET")
    secret = JiraSecret.model_validate(secret)

    with get_jira_client(secret) as client:
        offset = 0
        logs = []

        params = {}
        if start_date is not None:
            params["from"] = start_date
        if end_date is not None:
            params["to"] = end_date
        if filter is not None:
            if any(" " in f for f in filter):
                raise ValueError("Filter strings must not contain spaces.")
            params["filter"] = " ".join(filter)

        while limit is None or len(logs) < limit:
            params["offset"] = offset
            response = client.get("/auditing/record", params=params)
            response.raise_for_status()
            data = response.json()

            logs.extend(data["records"])
            offset = len(logs)
            if offset == data["total"] or is_empty(data["records"]):
                break

        return logs if limit is None else logs[:limit]


@action(
    display_name="Get Project",
    display_namespace="Jira",
    description="Get a Jira project",
    secrets_placeholders=["JIRA_SECRET"],
)
def get_jira_project(
    project_id_or_key: Annotated[
        str,
        ArgumentMetadata(
            display_name="Project ID or Key",
            description="The ID or the key of the project",
        ),
    ],
) -> dict[str, JsonValue]:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-projects/#api-rest-api-3-project-projectidorkey-get
    secret = ctx.get().secrets.get("JIRA_SECRET")
    secret = JiraSecret.model_validate(secret)

    with get_jira_client(secret) as client:
        response = client.get(f"/project/{project_id_or_key}")
        response.raise_for_status()
        return response.json()


@action(
    display_name="Get Transitions",
    display_namespace="Jira",
    description="Get a Jira transition",
    secrets_placeholders=["JIRA_SECRET"],
)
def get_jira_transitions(
    issue_id_or_key: Annotated[
        str,
        ArgumentMetadata(
            display_name="Issue ID or Key",
            description="The ID or the key of the issue",
        ),
    ],
) -> list[dict[str, JsonValue]]:
    # https://developer.atlassian.com/cloud/jira/platform/rest/v3/api-group-issues/#api-rest-api-3-issue-issueidorkey-transitions-get
    secret = ctx.get().secrets.get("JIRA_SECRET")
    secret = JiraSecret.model_validate(secret)

    with get_jira_client(secret) as client:
        response = client.get(f"/issue/{issue_id_or_key}/transitions")
        response.raise_for_status()
        return response.json()



from typing import Annotated
from httpx import Client
import base64
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.secret.secret import register_secret


@register_secret(secret_type="VirusTotal")
class VirusTotalSecret(BaseModel):
    api_key: str


def get_virus_total_client(secret: VirusTotalSecret) -> Client:
    return Client(
        base_url="https://www.virustotal.com/api/v3",
        headers={
            "x-apikey": secret.api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Analyze File Hash",
    display_namespace="VirusTotal",
    description="Analyze a hash using VirusTotal",
    secrets_placeholders=["VIRUS_TOTAL_SECRET"],
)
def virus_total_analyze_hash(
    hash: Annotated[
        str,
        ArgumentMetadata(
            display_name="Hash",
            description="The file hash to analyze",
        ),
    ],
) -> JsonValue:
    # https://docs.virustotal.com/reference/file-info
    secret = ctx.get().secrets.get("VIRUS_TOTAL_SECRET")
    secret = VirusTotalSecret.model_validate(secret)
    with get_virus_total_client(secret) as client:
        response = client.get(f"/files/{hash}")
        response.raise_for_status()
        return response.json()


@action(
    display_name="Analyze Domain",
    display_namespace="VirusTotal",
    description="Analyze a domain using VirusTotal",
    secrets_placeholders=["VIRUS_TOTAL_SECRET"],
)
def virus_total_analyze_domain(
    domain: Annotated[
        str,
        ArgumentMetadata(
            display_name="Domain",
            description="The domain to analyze",
        ),
    ],
) -> JsonValue:
    # https://docs.virustotal.com/reference/domain-info
    secret = ctx.get().secrets.get("VIRUS_TOTAL_SECRET")
    secret = VirusTotalSecret.model_validate(secret)
    with get_virus_total_client(secret) as client:
        response = client.get(f"/domains/{domain}")
        response.raise_for_status()
        return response.json()


@action(
    display_name="Analyze IP Address",
    display_namespace="VirusTotal",
    description="Analyze an IP address using VirusTotal",
    secrets_placeholders=["VIRUS_TOTAL_SECRET"],
)
def virus_total_analyze_ip(
    ip_address: Annotated[
        str,
        ArgumentMetadata(
            display_name="IP Address", description="The IP address to analyze"
        ),
    ],
) -> JsonValue:
    # https://developers.virustotal.com/reference/ip-addresses
    secret = ctx.get().secrets.get("VIRUS_TOTAL_SECRET")
    secret = VirusTotalSecret.model_validate(secret)
    with get_virus_total_client(secret) as client:
        response = client.get(f"/ip_addresses/{ip_address}")
        response.raise_for_status()
        return response.json()


@action(
    display_name="Analyze URL",
    display_namespace="VirusTotal",
    description="Analyze a URL using VirusTotal",
    secrets_placeholders=["VIRUS_TOTAL_SECRET"],
)
def virus_total_analyze_url(
    url: Annotated[
        str,
        ArgumentMetadata(
            display_name="URL",
            description="The URL to analyze",
        ),
    ],
) -> JsonValue:
    # https://docs.virustotal.com/reference/url-info
    secret = ctx.get().secrets.get("VIRUS_TOTAL_SECRET")
    secret = VirusTotalSecret.model_validate(secret)
    with get_virus_total_client(secret) as client:
        url_base64 = base64.b64encode(url.encode()).decode()
        url_base64 = url_base64.rstrip("=")
        response = client.get(f"/urls/{url_base64}")
        response.raise_for_status()
        return response.json()



from typing import Annotated
import json
import time
from datetime import datetime, timedelta, UTC
import yaml
from copy import deepcopy

from admyral.action import action, ArgumentMetadata
from admyral.typings import JsonValue
from admyral.context import ctx
from admyral.compiler.condition_compiler import compile_condition_str
from admyral.workers.if_condition_executor import (
    ConditionReferenceResolution,
    ConditionEvaluator,
)


@action(
    display_name="Deserialize JSON String",
    display_namespace="Admyral",
    description="Deserializes a JSON string.",
)
def deserialize_json_string(
    serialized_json: Annotated[
        str,
        ArgumentMetadata(
            display_name="Serialized JSON String",
            description="Deserialized JSON string.",
        ),
    ],
) -> JsonValue:
    return json.loads(serialized_json)


@action(
    display_name="Serialize JSON String",
    display_namespace="Admyral",
    description="Serializes a JSON string.",
)
def serialize_json_string(
    json_value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="JSON",
            description="A JSON value.",
        ),
    ],
) -> str:
    return json.dumps(json_value)


@action(
    display_name="Wait",
    display_namespace="Admyral",
    description="Waits for a specified number of seconds.",
)
def wait(
    seconds: Annotated[
        int,
        ArgumentMetadata(
            display_name="Seconds",
            description="Number of seconds to wait.",
        ),
    ],
) -> None:
    time.sleep(seconds)


@action(
    display_name="Transform",
    display_namespace="Admyral",
    description="Transforms a JSON.",
)
def transform(
    value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="Value",
            description="A JSON value.",
        ),
    ],
) -> JsonValue:
    return value


@action(
    display_name="Get Time Interval of Last N Hours",
    display_namespace="Admyral",
    description="Get the time interval of the last N hours. A list of length 2 containing the start and end time is returned.",
)
def get_time_interval_of_last_n_hours(
    n_hours: Annotated[
        int,
        ArgumentMetadata(
            display_name="Number of Hours",
            description="Number of hours to go back.",
        ),
    ] = 1,
) -> list[str]:
    end_time = datetime.now(UTC).replace(minute=0, second=0, microsecond=0)
    start_time = (
        (end_time - timedelta(hours=n_hours)).isoformat().replace("+00:00", "Z")
    )
    return [start_time, end_time.isoformat().replace("+00:00", "Z")]


@action(
    display_name="Get Time Interval of Last N Days",
    display_namespace="Admyral",
    description="Get the time interval of the last N days. A list of length 2 containing the start and end time is returned.",
)
def get_time_interval_of_last_n_days(
    n_days: Annotated[
        int,
        ArgumentMetadata(
            display_name="Number of Days",
            description="Number of days to go back.",
        ),
    ] = 1,
    only_date: Annotated[
        bool,
        ArgumentMetadata(
            display_name="Only Date",
            description="If true, only the date is returned.",
        ),
    ] = False,
) -> list[str]:
    end_time = datetime.now(UTC).replace(hour=0, minute=0, second=0, microsecond=0)
    start_time = end_time - timedelta(days=n_days)

    if only_date:
        return [start_time.strftime("%Y-%m-%d"), end_time.strftime("%Y-%m-%d")]

    start_time = start_time.isoformat().replace("+00:00", "Z")
    return [start_time, end_time.isoformat().replace("+00:00", "Z")]


@action(
    display_name="Format JSON to List View String",
    display_namespace="Admyral",
    description="Converts a JSON value to a list view string (YAML format).",
)
def format_json_to_list_view_string(
    json_value: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="JSON",
            description="A JSON value.",
        ),
    ],
) -> str:
    return yaml.dump(json_value, default_flow_style=False)


@action(
    display_name="Send to Workflow",
    display_namespace="Admyral",
    description="Sends data to a workflow.",
)
def send_to_workflow(
    workflow_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Workflow Name",
            description="The name of the workflow to send the data to.",
        ),
    ],
    payload: Annotated[
        JsonValue,
        ArgumentMetadata(
            display_name="Payload",
            description="Payload to send to the workflow.",
        ),
    ],
) -> None:
    ctx.get().send_to_workflow_sync(workflow_name, payload)


# TODO: remove with the introduction of for-loops
@action(
    display_name="Send List Elements to a Workflow",
    display_namespace="Admyral",
    description="WARNING: This action is temporary and will be deprecated as soon as For Loops are released. Sends a list of elements to a workflow.",
)
def send_list_elements_to_workflow(
    workflow_name: Annotated[
        str,
        ArgumentMetadata(
            display_name="Workflow Name",
            description="The name of the workflow to send the elements to.",
        ),
    ],
    elements: Annotated[
        list[JsonValue],
        ArgumentMetadata(
            display_name="Elements",
            description="A list of elements to send to the workflow.",
        ),
    ],
    shared_data: Annotated[
        JsonValue | None,
        ArgumentMetadata(
            display_name="Shared Data",
            description="Shared data to send to the workflow.",
        ),
    ] = None,
) -> None:
    for element in elements:
        send_to_workflow(
            workflow_name=workflow_name,
            payload={
                "element": element,
                "shared": shared_data,
            },
        )


@action(
    display_name="Split Text",
    display_namespace="Admyral",
    description="Splits a text into a list of strings.",
)
def split_text(
    text: Annotated[
        str, ArgumentMetadata(display_name="Text", description="The text to split.")
    ],
    pattern: Annotated[
        str,
        ArgumentMetadata(
            display_name="Pattern", description="The pattern to split by."
        ),
    ],
) -> list[str]:
    return text.split(pattern)


@action(
    display_name="Build Lookup Table",
    display_namespace="Admyral",
    description="Builds a lookup table from an array of objects.",
)
def build_lookup_table(
    input_list: Annotated[
        list[JsonValue],
        ArgumentMetadata(
            display_name="Input List",
            description="A list of objects to build the lookup table from.",
        ),
    ],
    key_path: Annotated[
        list[str],
        ArgumentMetadata(
            display_name="Key Path",
            description='The key path to use as the lookup key (e.g., ["email"]).',
        ),
    ],
) -> dict[str, JsonValue]:
    lookup_table = {}
    for obj in input_list:
        key = obj
        for key_part in key_path:
            key = key[key_part]
        lookup_table[key] = obj
    return lookup_table


@action(
    display_name="Filter",
    display_namespace="Admyral",
    description="Filters a list of objects by a condition. The action iterates over the list of objects and applies a filter to each object. "
    "When the condition evaluates to false, then the object is removed from the list. The current object in the filter condition is represented by 'x'. "
    "In order to use results from previous actions in the filter condition, you can use the 'values' parameter.",
)
def filter(
    input_list: Annotated[
        list[JsonValue],
        ArgumentMetadata(
            display_name="Input List",
            description="The list of objects to filter.",
        ),
    ],
    filter: Annotated[
        str,
        ArgumentMetadata(
            display_name="Filter",
            description="The filter condition (e.g., \"x['email'] not in okta_users\"). Note: The current object is represented by 'x'.",
        ),
    ],
    values: Annotated[
        dict[str, JsonValue],
        ArgumentMetadata(
            display_name="Values",
            description='Values (results from previous actions) to use in the filter condition for comparisons (e.g., {"okta_users": okta_users}).',
        ),
    ],
) -> list[JsonValue]:
    compiled_filter_expr = compile_condition_str(filter)
    filtered_input_list = []
    for x in input_list:
        values["x"] = x
        copy_compiled_filter_expr = deepcopy(compiled_filter_expr)
        expr = ConditionReferenceResolution(values).resolve_references(
            copy_compiled_filter_expr
        )
        if ConditionEvaluator().evaluate(expr):
            filtered_input_list.append(x)
    return filtered_input_list


@action(
    display_name="Join Lists",
    display_namespace="Admyral",
    description="Performs an inner join between 2 lists, i.e., returns a list of objects that have matching keys in both lists.",
)
def join_lists(
    list1: Annotated[
        list[JsonValue],
        ArgumentMetadata(
            display_name="List 1",
            description="The first list to join.",
        ),
    ],
    list1_join_key_paths: Annotated[
        list[list[str]],
        ArgumentMetadata(
            display_name="Keys",
            description="The keys to join the lists on.",
        ),
    ],
    list2: Annotated[
        list[JsonValue],
        ArgumentMetadata(
            display_name="List 2",
            description="The second list to join.",
        ),
    ],
    list2_join_key_paths: Annotated[
        list[list[str]],
        ArgumentMetadata(
            display_name="Keys",
            description="The keys to join the lists on.",
        ),
    ],
    key_prefix_list1: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Key Prefix",
            description="The prefix to add to the keys of the second list.",
        ),
    ] = None,
    key_prefix_list2: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Key Prefix",
            description="The prefix to add to the keys of the second list.",
        ),
    ] = None,
) -> list[JsonValue]:
    def calculate_join_key(entry, key_paths):
        composed_key = []
        for key_path in key_paths:
            key = entry
            for key_part in key_path:
                key = key[key_part]
            composed_key.append(key)
        return tuple(composed_key)

    ht = {}
    for entry in list2:
        key = calculate_join_key(entry, list2_join_key_paths)
        ht[key] = entry

    join_result = []
    for entry in list1:
        key = calculate_join_key(entry, list1_join_key_paths)
        if entry_list2 := ht.get(key):
            entry_list2 = deepcopy(entry_list2)
            new_entry = deepcopy(entry)
            if key_prefix_list1:
                new_entry = {f"{key_prefix_list1}{k}": v for k, v in new_entry.items()}
            if key_prefix_list2:
                entry_list2 = {
                    f"{key_prefix_list2}{k}": v for k, v in entry_list2.items()
                }
            new_entry |= entry_list2

            join_result.append(new_entry)

    return join_result


@action(
    display_name="Select Fields from Objects in List",
    display_namespace="Admyral",
    description="Selects fields from a list of JSON objects.",
)
def select_fields_from_objects_in_list(
    input_list: Annotated[
        list[dict[str, JsonValue]],
        ArgumentMetadata(
            display_name="Input List",
            description="The list of objects to select fields from.",
        ),
    ],
    fields: Annotated[
        list[str],
        ArgumentMetadata(
            display_name="Fields",
            description="The keys to select from the objects.",
        ),
    ],
) -> list[dict[str, JsonValue]]:
    return [{field: x[field] for field in fields} for x in input_list]



import subprocess
import json
import os
import threading

from admyral.logger import get_logger
from admyral.exceptions import NonRetryableActionError


logger = get_logger(__name__)


# Note: steampipe currently does not support concurrent queries
# because it launches a server process with an embedded postgres
# instance for each steampipe query execution and each concurrent
# steampipe query execution would connect to the same server instance.
STEAMPIPE_LOCK = threading.Lock()


def _get_steampipe_executable() -> str:
    if os.path.exists("/usr/local/bin/steampipe"):
        return "/usr/local/bin/steampipe"
    if os.path.exists("/opt/homebrew/bin/steampipe"):
        return "/opt/homebrew/bin/steampipe"
    raise ValueError("Steampipe installation not found.")


def run_steampipe_query(
    query: str,
    aws_access_key_id: str,
    aws_secret_access_key: str,
) -> dict:
    env = os.environ.copy()  # Start with the current environment
    env["AWS_ACCESS_KEY_ID"] = aws_access_key_id
    env["AWS_SECRET_ACCESS_KEY"] = aws_secret_access_key

    if "AWS_ACCOUNT_ID" in env:
        del env["AWS_ACCOUNT_ID"]
    if "AWS_DEFAULT_REGION" in env:
        del env["AWS_DEFAULT_REGION"]

    # Disable caching
    env["STEAMPIPE_CACHE_PATH"] = "false"

    try:
        steampipe_executable = _get_steampipe_executable()
        with STEAMPIPE_LOCK:
            result = subprocess.run(
                [steampipe_executable, "query", query, "--output", "json"],
                capture_output=True,
                text=True,
                check=True,
                env=env,
            )
        return json.loads(result.stdout)
    except subprocess.CalledProcessError as e:
        logger.error(
            f"An error occurred while executing the steampipe query. Error: {str(e)}. Stderr: {e.stderr}"
        )
        raise NonRetryableActionError(
            f"An error occurred while executing the query. Error: {str(e)}. Stderr: {e.stderr}"
        )



from fastapi import APIRouter, status, Depends
from pydantic import BaseModel

from admyral.server.auth import authenticate
from admyral.models import AuthenticatedUser, ApiKey
from admyral.server.deps import get_admyral_store
from admyral.utils.api_key import generate_api_key


router = APIRouter()


@router.get("", status_code=status.HTTP_200_OK)
async def list_api_keys(
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> list[ApiKey]:
    return await get_admyral_store().list_api_keys(authenticated_user.user_id)


class CreateApiKeyRequest(BaseModel):
    name: str


class CreateApiKeyResponse(BaseModel):
    key: ApiKey
    secret: str


@router.post("", status_code=status.HTTP_201_CREATED)
async def create_api_key(
    request: CreateApiKeyRequest,
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> CreateApiKeyResponse:
    secret_key = generate_api_key()
    api_key = await get_admyral_store().store_api_key(
        authenticated_user.user_id, request.name, secret_key
    )
    return CreateApiKeyResponse(key=api_key, secret=secret_key)


@router.delete("/{api_key_id}", status_code=status.HTTP_204_NO_CONTENT)
async def delete_api_key(
    api_key_id: str,
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> None:
    await get_admyral_store().delete_api_key(authenticated_user.user_id, api_key_id)



from typing import TYPE_CHECKING, TypeVar, Callable, Any
from temporalio import activity
import inspect
from uuid import uuid4
import time

from admyral.context import ExecutionContext
from admyral.utils.json import throw_if_not_allowed_return_type
from admyral.utils.future_executor import execute_future
from admyral.logger import get_logger
from admyral.context import ctx
from admyral.secret.secrets_access import Secrets, SecretsStoreAccessImpl
from admyral.workers.shared_worker_state import SharedWorkerState
from admyral.typings import JsonValue
from admyral.exceptions import NonRetryableActionError

if TYPE_CHECKING:
    F = TypeVar("F", bound=Callable[..., Any])

logger = get_logger(__name__)


def action_executor(action_type: str, func: "F") -> "F":
    if inspect.iscoroutinefunction(func):

        @activity.defn(name=action_type)
        async def async_execute(
            ctx_dict: dict, secret_mappings: dict[str, str], args: dict[str, Any]
        ) -> tuple[str, Any]:
            logger.info(f"Executing async action: {action_type}")

            exec_ctx = ExecutionContext(**ctx_dict)
            exec_ctx.step_id = str(uuid4())

            try:
                exec_ctx.secrets = Secrets(
                    SecretsStoreAccessImpl(
                        user_id=exec_ctx.user_id,
                        secret_mappings=secret_mappings,
                        secrets_manager=SharedWorkerState.get_secrets_manager(),
                    )
                )

                # make available to the function as a contextvar
                ctx.set(exec_ctx)
                # for wait actions, we skip the function call because the waiting is already executed in the WorkflowExecutor
                if exec_ctx.action_type != "wait":
                    logger.info(f"Executing {action_type}...")
                    start = time.time()
                    result = await func(**args)
                    end = time.time()
                    logger.info(
                        f"Finished execution of {action_type} (async) in {int((end - start) * 1000)}ms"
                    )
                else:
                    result = None
                throw_if_not_allowed_return_type(result)
            except Exception as e:
                # Store error
                await _store_action_error(exec_ctx, str(e), args)
                raise NonRetryableActionError(str(e))

            await _store_action_result(exec_ctx, result, args)

            return exec_ctx.step_id, result

        return async_execute

    else:

        @activity.defn(name=action_type)
        def sync_execute(
            ctx_dict: dict, secret_mappings: dict[str, str], args: dict[str, Any]
        ) -> tuple[str, Any]:
            logger.info(f"Executing sync action: {action_type}")

            exec_ctx = ExecutionContext(**ctx_dict)
            exec_ctx.step_id = str(uuid4())

            try:
                exec_ctx.secrets = Secrets(
                    SecretsStoreAccessImpl(
                        user_id=exec_ctx.user_id,
                        secret_mappings=secret_mappings,
                        secrets_manager=SharedWorkerState.get_secrets_manager(),
                    )
                )

                # make available to the function as a contextvar
                ctx.set(exec_ctx)

                # for wait actions, we skip the function call because the waiting is already executed in the WorkflowExecutor
                if exec_ctx.action_type != "wait":
                    logger.info(f"Executing {action_type}...")
                    start = time.time()
                    result = func(**args)
                    end = time.time()
                    logger.info(
                        f"Finished execution of {action_type} (sync) in {int((end - start) * 1000)}ms"
                    )
                else:
                    result = None
                throw_if_not_allowed_return_type(result)
            except Exception as e:
                # Store error
                execute_future(_store_action_error(exec_ctx, str(e), args))
                raise NonRetryableActionError(str(e))

            execute_future(_store_action_result(exec_ctx, result, args))

            return exec_ctx.step_id, result

        return sync_execute


async def _store_action_result(
    exec_ctx: ExecutionContext, result: JsonValue, args: dict[str, Any]
) -> None:
    await SharedWorkerState.get_store().store_action_result(
        exec_ctx.step_id,
        exec_ctx.run_id,
        exec_ctx.action_type,
        exec_ctx.prev_step_id,
        result,
        args,
    )


async def _store_action_error(
    exec_ctx: ExecutionContext, error: str, args: dict[str, Any]
) -> None:
    await SharedWorkerState.get_store().store_workflow_run_error(
        exec_ctx.step_id,
        exec_ctx.run_id,
        exec_ctx.action_type,
        exec_ctx.prev_step_id,
        error,
        args,
    )



from fastapi import APIRouter, status, HTTPException, Depends
from fastapi.responses import StreamingResponse
import json
import io

from admyral.server.auth import authenticate
from admyral.models import (
    AuthenticatedUser,
    WorkflowRunMetadata,
    WorkflowRunStepMetadata,
    WorkflowRunStepWithSerializedResult,
)
from admyral.server.deps import get_admyral_store


router = APIRouter()


MAX_SERIALIZED_RESULT_LENGTH = 10_000


@router.get("/{workflow_id}", status_code=status.HTTP_200_OK)
async def list_workflow_runs(
    workflow_id: str, authenticated_user: AuthenticatedUser = Depends(authenticate)
) -> list[WorkflowRunMetadata]:
    """
    List all workflow runs.

    Args:
        workflow_id: The workflow id

    Returns:
        A list of workflow runs.
    """
    return await get_admyral_store().list_workflow_runs(
        authenticated_user.user_id, workflow_id
    )


@router.get("/{workflow_id}/{workflow_run_id}", status_code=status.HTTP_200_OK)
async def get_workflow_run_steps(
    workflow_id: str,
    workflow_run_id: str,
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> list[WorkflowRunStepMetadata]:
    """
    Get workflow runs by workflow id.

    Args:
        workflow_id: The workflow id
        workflow_run_id: The workflow run id

    Returns:
        A list of workflow run steps.
    """
    return await get_admyral_store().list_workflow_run_steps(
        authenticated_user.user_id, workflow_id, workflow_run_id
    )


@router.get(
    "/{workflow_id}/{workflow_run_id}/{workflow_run_step_id}",
    status_code=status.HTTP_200_OK,
)
async def get_workflow_run_step(
    workflow_id: str,
    workflow_run_id: str,
    workflow_run_step_id: str,
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> WorkflowRunStepWithSerializedResult:
    """
    Get workflow run step.

    Args:
        workflow_id: The workflow id
        workflow_run_id: The workflow run id
        workflow_run_step_id: The workflow run step id

    Returns:
        The workflow run step.
    """
    step = await get_admyral_store().get_workflow_run_step(
        authenticated_user.user_id, workflow_id, workflow_run_id, workflow_run_step_id
    )
    if not step:
        raise HTTPException(status_code=404, detail="Step not found")

    serialized_result = json.dumps(step.result, indent=4)
    if len(serialized_result) > MAX_SERIALIZED_RESULT_LENGTH:
        serialized_result = (
            serialized_result[:MAX_SERIALIZED_RESULT_LENGTH]
            + "...\nResult truncated because it is too long."
        )
    return WorkflowRunStepWithSerializedResult.model_validate(
        {
            **step.model_dump(),
            "result": serialized_result,
        }
    )


@router.get(
    "/{workflow_id}/{workflow_run_id}/{workflow_run_step_id}/download",
    status_code=status.HTTP_200_OK,
)
async def download_workflow_run_step_result(
    workflow_id: str,
    workflow_run_id: str,
    workflow_run_step_id: str,
    authenticated_user: AuthenticatedUser = Depends(authenticate),
) -> StreamingResponse:
    """
    Download the full result of a workflow run step.

    Args:
        workflow_id: The workflow id
        workflow_run_id: The workflow run id
        workflow_run_step_id: The workflow run step id

    Returns:
        The complete result as a downloadable JSON file.
    """
    step = await get_admyral_store().get_workflow_run_step(
        authenticated_user.user_id, workflow_id, workflow_run_id, workflow_run_step_id
    )
    if not step:
        raise HTTPException(status_code=404, detail="Step not found")

    # Convert to JSON and create a string buffer
    json_str = json.dumps(step.result, indent=4)
    json_bytes = json_str.encode("utf-8")

    return StreamingResponse(
        io.BytesIO(json_bytes),
        media_type="application/json",
        headers={
            "Content-Disposition": f'attachment; filename="step_result_{workflow_run_step_id}.json"'
        },
    )



from __future__ import annotations
from pydantic import BaseModel
from datetime import datetime

from admyral.typings import JsonValue


class WorkflowRunStep(BaseModel):
    step_id: str
    action_type: str
    prev_step_id: str | None = None
    logs: str | None = None
    result: JsonValue | None = None
    error: str | None = None
    input_args: JsonValue | None = None


class WorkflowRunStepWithSerializedResult(BaseModel):
    step_id: str
    action_type: str
    prev_step_id: str | None = None
    logs: str | None = None
    result: str | None = None
    error: str | None = None
    input_args: JsonValue | None = None


class WorkflowRun(BaseModel):
    run_id: str
    trigger_id: int
    payload: JsonValue | None
    created_at: datetime
    completed_at: datetime | None = None
    failed_at: datetime | None = None
    canceled_at: datetime | None = None
    steps: list[WorkflowRunStep] | None = None


class WorkflowRunMetadata(BaseModel):
    run_id: str
    created_at: datetime
    completed_at: datetime | None = None
    failed_at: datetime | None = None
    canceled_at: datetime | None = None


class WorkflowRunStepMetadata(BaseModel):
    step_id: str
    action_type: str
    error: str | None = None



from typing import Annotated
from httpx import Client
from datetime import timedelta
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.utils.time import utc_now
from admyral.secret.secret import register_secret


@register_secret(secret_type="Wiz")
class WizSecret(BaseModel):
    client_id: str
    client_secret: str
    auth_url: str
    api_endpoint: str


def _fetch_token(
    client_id: str,
    client_secret: str,
    auth_url: str,
) -> str:
    with Client(
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
    ) as client:
        response = client.post(
            auth_url,
            json={
                "client_id": client_id,
                "client_secret": client_secret,
                "audience": "wiz-api",
                "grant_type": "client_credentials",
            },
        )
        response.raise_for_status()
        return response.json()["access_token"]


def get_wiz_client(secret: WizSecret) -> Client:
    access_token = _fetch_token(secret.client_id, secret.client_secret, secret.auth_url)
    return Client(
        base_url=secret.api_endpoint,
        headers={
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {access_token}",
        },
    )


_LIST_ALERTS_QUERY = """
query IssuesTable($filterBy: IssueFilters, $first: Int, $after: String, $orderBy: IssueOrder) {
    issues: issuesV2(filterBy: $filterBy, first: $first, after: $after, orderBy: $orderBy) {
        nodes {
            id
            control {
                id
                name
                description
                resolutionRecommendation
                securitySubCategories {
                    title
                    category {
                        name
                        framework {
                            name
                        }
                    }
                }
            }
            createdAt
            updatedAt
            sourceRule {
                id
                name
            }
            dueAt
            resolvedAt
            statusChangedAt
            project {
                id
                name
                slug
                businessUnit
                riskProfile {
                    businessImpact
                }
            }
            status
            severity
            type
            entitySnapshot {
                id
                type
                nativeType
                name
                status
                cloudPlatform
                cloudProviderURL
                providerId
                region
                resourceGroupExternalId
                subscriptionExternalId
                subscriptionName
                subscriptionTags
                tags
                externalId
            }
            notes {
                createdAt
                updatedAt
                text
                user {
                    name
                    email
                }
                serviceAccount {
                    name
                }
            }
            serviceTickets {
                externalId
                name
                url
            }
        }
        pageInfo {
            hasNextPage
            endCursor
        }
    }
}
"""


@action(
    display_name="List Alerts",
    display_namespace="Wiz",
    description="List alerts from Wiz. If no time range is provided, the alerts from the last 24 hours will be listed.",
    secrets_placeholders=["WIZ_SECRET"],
)
def list_wiz_alerts(
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of cases to list.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # https://github.com/criblio/collector-templates/blob/main/collectors/rest/wiz/collector-wiz-issues.json

    secret = ctx.get().secrets.get("WIZ_SECRET")
    secret = WizSecret.model_validate(secret)

    with get_wiz_client(secret) as client:
        alerts = []

        end_cursor = None
        variables = {
            "filterBy": {
                "sourceRule": {"id": []},
                "relatedEntity": {"type": []},
                "status": ["OPEN", "IN_PROGRESS", "RESOLVED", "REJECTED"],
                "severity": ["INFORMATIONAL", "LOW", "MEDIUM", "HIGH", "CRITICAL"],
                "type": [
                    "TOXIC_COMBINATION",
                    "THREAT_DETECTION",
                    "CLOUD_CONFIGURATION",
                ],
                "createdAt": {
                    "after": start_time
                    if start_time
                    else (utc_now() - timedelta(days=1)).strftime("%Y-%m-%dT%H:%M:%SZ"),
                    "before": end_time
                    if end_time
                    else utc_now().strftime("%Y-%m-%dT%H:%M:%SZ"),
                },
            },
            "first": 100,
        }

        while len(alerts) < limit:
            response = client.post(
                "",
                json={
                    "query": _LIST_ALERTS_QUERY,
                    "variables": variables,
                },
            )
            response.raise_for_status()
            data = response.json()["data"]["issues"]

            alerts.extend(data.get("nodes", []))

            if not data["pageInfo"]["hasNextPage"]:
                break

            if end_cursor := data["pageInfo"].get("endCursor"):
                variables["after"] = end_cursor
            else:
                break

        return alerts[:limit]



from admyral.actions.integrations.cases.jira import (
    create_jira_issue,
    update_jira_issue_status,
    comment_jira_issue_status,
    search_jira_issues,
    get_jira_audit_records,
    get_jira_project,
    get_jira_transitions,
)
from admyral.actions.integrations.cases.opsgenie import create_opsgenie_alert
from admyral.actions.integrations.cases.pagerduty import create_pagerduty_incident

__all__ = [
    "create_jira_issue",
    "update_jira_issue_status",
    "comment_jira_issue_status",
    "search_jira_issues",
    "get_jira_audit_records",
    "create_opsgenie_alert",
    "create_pagerduty_incident",
    "get_jira_project",
    "get_jira_transitions",
]



import click
import yaml

from admyral.cli.cli import cli
from admyral.config.config import get_user_config_file


@cli.command("connect", help="Connect to remote Admyral server.")
@click.argument("base_url", type=str)
@click.argument("api_key", type=str)
def connect(base_url: str, api_key: str):
    """Connect to the Admyral server."""
    with open(get_user_config_file(), "r") as f:
        config = yaml.safe_load(f)
    config["cli_target"] = base_url
    config["api_key"] = api_key
    with open(get_user_config_file(), "w") as f:
        yaml.dump(config, f)


@cli.command("disconnect", help="Disconnect from remote Admyral server.")
def disconnect():
    """Disconnect from the Admyral server."""
    with open(get_user_config_file(), "r") as f:
        config = yaml.safe_load(f)
    if "cli_target" in config:
        del config["cli_target"]
    if "api_key" in config:
        del config["api_key"]
    with open(get_user_config_file(), "w") as f:
        yaml.dump(config, f)



from typing import Annotated
from googleapiclient.discovery import build
from google.oauth2.service_account import Credentials as ServiceAccountCredentials
from difflib import ndiff
import requests
from dateutil import parser
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.utils.collections import is_empty
from admyral.secret.secret import register_secret


@register_secret(secret_type="Google Drive")
class GoogleDriveSecret(BaseModel):
    type: str
    project_id: str
    private_key_id: str
    private_key: str
    client_email: str
    auth_uri: str
    token_uri: str
    auth_provider_x509_cert_url: str
    client_x509_cert_url: str
    universe_domain: str


@retry(
    stop=stop_after_attempt(10),
    wait=wait_exponential(multiplier=1, min=1, max=10),
    retry=retry_if_exception_type(requests.HTTPError),
    reraise=True,
)
def _export_google_docs_revision(
    export_link: str, mime_type: str, creds: ServiceAccountCredentials
) -> requests.Response:
    """
    For very recent revisions, we might need to retry the export a couple of times since
    there is a slight delay between the revision being created and the export being available.
    """
    export_response = requests.get(
        export_link,
        headers={"Authorization": f"Bearer {creds.token}", "Accept": mime_type},
    )
    export_response.raise_for_status()
    return export_response


@action(
    display_name="List Google Docs Revisions",
    display_namespace="Google Drive",
    description="Fetch revisions of a Google Docs document.",
    secrets_placeholders=["GOOGLE_DRIVE_SECRET"],
)
def list_google_docs_revisions(
    file_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="File ID", description="The ID of the Google Doc."
        ),
    ],
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the revisions to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the revisions to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
) -> list[dict[str, JsonValue]]:
    # https://googleapis.github.io/google-api-python-client/docs/dyn/drive_v3.revisions.html#list

    secret = ctx.get().secrets.get("GOOGLE_DRIVE_SECRET")
    secret = GoogleDriveSecret.model_validate(secret)
    creds = ServiceAccountCredentials.from_service_account_info(
        info=secret, scopes=["https://www.googleapis.com/auth/drive.readonly"]
    )

    drive_service = build("drive", "v3", credentials=creds)

    if start_time:
        start_time = parser.parse(start_time)
    if end_time:
        end_time = parser.parse(end_time)

    result = []

    revision_request = drive_service.revisions().list(
        fileId=file_id,
        fields="nextPageToken, revisions(id, modifiedTime, lastModifyingUser, exportLinks)",
    )

    while revision_request is not None:
        revision_response = revision_request.execute()
        revisions = revision_response.get("revisions", [])

        for revision in revisions:
            revision_id = revision["id"]
            modified_time = parser.parse(revision["modifiedTime"])

            if (start_time and modified_time < start_time) or (
                end_time and end_time < modified_time
            ):
                # Skip revisions outside the time range
                continue

            mime_type = "text/plain"
            export_link = revision["exportLinks"][mime_type]
            export_response = _export_google_docs_revision(
                export_link, mime_type, creds
            )

            prev_text = "" if is_empty(result) else result[-1]["content"]
            diff = "\n".join(
                line
                for line in ndiff(
                    prev_text.splitlines(),
                    export_response.text.splitlines(),
                )
                if line.startswith("+ ")
                or line.startswith("- ")
                or line.startswith("? ")
            )

            result.append(
                {
                    "id": revision_id,
                    "lastModifyingUser": revision["lastModifyingUser"]["emailAddress"],
                    "modifiedTime": revision["modifiedTime"],
                    "content": export_response.text,
                    "diff": diff,
                }
            )

        revision_request = drive_service.revisions().list_next(
            revision_request, revision_response
        )

    return result


@action(
    display_name="List Google Drive Files with Link Sharing Enabled",
    display_namespace="Google Drive",
    description="List all files in a Google Drive of an organization which have public link sharing enabled.",
    secrets_placeholders=["GOOGLE_DRIVE_SECRET"],
)
def list_google_drive_files_with_link_sharing_enabled(
    customer_id: Annotated[
        str,
        ArgumentMetadata(
            display_name="Customer ID",
            description="The customer ID of your Google Workspace.",
        ),
    ],
    admin_email: Annotated[
        str,
        ArgumentMetadata(
            display_name="Admin Email",
            description="The email of an admin user for your Google Workspace for delegated access.",
        ),
    ],
    limit: Annotated[
        int | None,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of files to return. If not specified, all files are returned.",
        ),
    ] = 100,
) -> list[dict[str, JsonValue]]:
    secret = ctx.get().secrets.get("GOOGLE_DRIVE_SECRET")
    secret = GoogleDriveSecret.model_validate(secret)
    creds = ServiceAccountCredentials.from_service_account_info(
        info=secret,
        scopes=[
            "https://www.googleapis.com/auth/drive.readonly",
            "https://www.googleapis.com/auth/admin.directory.user.readonly",
        ],
    )

    admin_delegated_creds = creds.with_subject(admin_email)

    admin_service = build("admin", "directory_v1", credentials=admin_delegated_creds)

    # make this a dict because we want to deduplciate the files
    all_public_files = {}

    # https://googleapis.github.io/google-api-python-client/docs/dyn/admin_directory_v1.html
    # https://developers.google.com/admin-sdk/directory/reference/rest/v1/users/list
    users_request = admin_service.users().list(customer=customer_id)
    while users_request is not None and (
        limit is None or len(all_public_files) < limit
    ):
        user_response = users_request.execute()

        # extract the publicly shared files for each user - for each user we look at the files which
        # the user has access to
        for user in user_response.get("users", []):
            user_email = user["primaryEmail"]

            # create delegated credentials for the user
            user_delegated_creds = creds.with_subject(user_email)
            drive_service = build("drive", "v3", credentials=user_delegated_creds)

            # extracts the files
            # https://googleapis.github.io/google-api-python-client/docs/dyn/drive_v3.files.html
            # https://developers.google.com/drive/api/reference/rest/v3/files/list
            list_files_request = drive_service.files().list(
                # spaces='drive',
                fields="nextPageToken, files(id, name, webViewLink, permissions, mimeType, modifiedTime, sharingUser, owners)",
                supportsAllDrives=True,
                includeItemsFromAllDrives=True,
                corpora="user",
                q="mimeType != 'application/vnd.google-apps.folder' and visibility = 'anyoneWithLink'",  # Exclude folders and include only files with public link sharing
            )
            while list_files_request is not None and (
                limit is None or len(all_public_files) < limit
            ):
                files_response = list_files_request.execute()
                for file in files_response.get("files", []):
                    all_public_files[file["id"]] = file
                list_files_request = drive_service.files().list_next(
                    list_files_request, files_response
                )

        users_request = admin_service.users().list_next(users_request, user_response)

    all_public_files = list(all_public_files.values())
    return all_public_files if limit is None else all_public_files[:limit]



from pydantic import BaseModel
from datetime import datetime


class AuthenticatedUser(BaseModel):
    user_id: str
    email: str


class User(BaseModel):
    id: str
    name: str | None
    email: str
    email_verified: datetime | None
    image: str | None
    created_at: datetime
    updated_at: datetime


class UserProfile(BaseModel):
    email: str



from typing import Annotated
from httpx import Client
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.secret.secret import register_secret


@register_secret(secret_type="AlienVault OTX")
class AlienVaultOTXSecret(BaseModel):
    api_key: str


def get_alienvault_otx_client(secret: AlienVaultOTXSecret) -> Client:
    return Client(
        base_url="https://otx.alienvault.com/api/v1",
        headers={
            "X-OTX-API-KEY": secret.api_key,
            "Content-Type": "application/json",
            "Accept": "application/json",
        },
    )


@action(
    display_name="Analyze Domain",
    display_namespace="AlienVault OTX",
    description="Analyze a domain using AlienVault OTX",
    secrets_placeholders=["ALIENVAULT_OTX_SECRET"],
)
def alienvault_otx_analyze_domain(
    domain: Annotated[
        str,
        ArgumentMetadata(
            display_name="Domain",
            description="The domain to analyze",
        ),
    ],
) -> JsonValue:
    # https://otx.alienvault.com/assets/static/external_api.html
    secret = ctx.get().secrets.get("ALIENVAULT_OTX_SECRET")
    secret = AlienVaultOTXSecret.model_validate(secret)

    with get_alienvault_otx_client(secret) as client:
        response = client.get(f"/indicators/domain/{domain}/general")
        response.raise_for_status()
        return response.json()



import secrets
import base64


def generate_api_key(length=48):
    """Generate a secure API key."""
    random_bytes = secrets.token_bytes(length)
    return base64.urlsafe_b64encode(random_bytes).decode("utf-8")



from abc import abstractmethod
import json

from admyral.db.store_interface import StoreInterface
from admyral.utils.crypto import decrypt_secret, encrypt_secret
from admyral.models import Secret, SecretMetadata
from admyral.config.config import CONFIG, SecretsManagerType
from admyral.utils.collections import is_empty
from admyral.secret.secret_registry import SecretRegistry


class SecretsManager:
    @abstractmethod
    async def get(self, user_id: str, secret_id: str) -> Secret | None:
        """
        Retrieve a secret for a user. If the secret does not exist, return None.

        Args:
            user_id: The user id of the user whose secret is retrieved.
            secret_id: The id of the secret to retrieve.
        """
        raise NotImplementedError("get method not implemented")

    @abstractmethod
    async def update(self, user_id: str, delta_secret: Secret) -> SecretMetadata:
        """
        Update a secret for a user. If the secret does not exist, raise a ValueError.

        Fields with an empty value are ignored and not updated.

        Args:
            user_id: The user id of the user whose secret is updated.
            delta_secret: The secret to update.
        """
        raise NotImplementedError("update method not implemented")

    @abstractmethod
    async def set(self, user_id: str, secret: Secret) -> SecretMetadata:
        """
        Set a secret for a user. If the secret already exists, it is overwritten.
        The function takes care of encrypting the secret.

        Args:
            user_id: The user id of the user whose secret is set.
            secret: The secret to set.
        """
        raise NotImplementedError("set method not implemented")

    @abstractmethod
    async def delete(self, user_id: str, secret_id: str) -> None:
        """
        Delete a secret for a user.

        Args:
            user_id: The user id of the user whose secret is deleted.
            secret_id: The id of the secret to delete.
        """
        raise NotImplementedError("delete method not implemented")

    @abstractmethod
    async def list(self, user_id: str) -> list[SecretMetadata]:
        """
        List the secret names for a user.

        Args:
            user_id: The user id of the user whose secrets are listed.
        """
        raise NotImplementedError("list method not implemented")


class SQLSecretsManager(SecretsManager):
    def __init__(self, db: StoreInterface) -> None:
        self.db = db

    async def get(self, user_id: str, secret_id: str) -> Secret | None:
        encrypted_secret = await self.db.get_secret(user_id, secret_id)
        return (
            Secret(
                secret_id=secret_id,
                secret=json.loads(decrypt_secret(encrypted_secret.encrypted_secret)),
            )
            if encrypted_secret
            else None
        )

    def _encrypt_secret(self, secret: dict[str, str]) -> str:
        return encrypt_secret(json.dumps(secret))

    async def update(self, user_id: str, delta_secret: Secret) -> SecretMetadata:
        if SecretRegistry.is_registered(delta_secret.secret_type):
            SecretRegistry.validate_schema(
                delta_secret.secret_type, delta_secret.secret
            )

        encrypted_secret = await self.db.get_secret(user_id, delta_secret.secret_id)
        if not encrypted_secret:
            raise ValueError(f"Secret {delta_secret.secret_id} does not exist.")

        if encrypted_secret.secret_type != delta_secret.secret_type:
            raise ValueError("Secret type cannot be changed.")

        secret = json.loads(decrypt_secret(encrypted_secret.encrypted_secret))

        for key, value in delta_secret.secret.items():
            # ignore empty values - empty values in the delta mean that no
            # update was provided for the key-value pair.
            if is_empty(value):
                continue
            secret[key] = value

        # remove the keys which are not present
        # we only allow editing the schema if the secret type is not valid, i.e.,
        # the secret is a custom secret.
        if not SecretRegistry.is_registered(delta_secret.secret_type):
            removal_keys = set(secret.keys()) - set(delta_secret.secret.keys())
            for key in removal_keys:
                secret.pop(key)

        secret_schema = list(secret.keys())
        updated_encrypted_secret = self._encrypt_secret(secret)

        return await self.db.compare_and_swap_secret(
            user_id,
            delta_secret.secret_id,
            encrypted_secret.encrypted_secret,
            updated_encrypted_secret,
            secret_schema,
            delta_secret.secret_type,
        )

    async def set(self, user_id: str, secret: Secret) -> SecretMetadata:
        if SecretRegistry.is_registered(secret.secret_type):
            SecretRegistry.validate_schema(secret.secret_type, secret.secret)

        encrypted_secret = self._encrypt_secret(secret.secret)
        secret_schema = list(secret.secret.keys())
        return await self.db.store_secret(
            user_id,
            secret.secret_id,
            encrypted_secret,
            secret_schema,
            secret.secret_type,
        )

    async def delete(self, user_id: str, secret_id: str) -> None:
        return await self.db.delete_secret(user_id, secret_id)

    async def list(self, user_id: str) -> list[SecretMetadata]:
        return await self.db.list_secrets(user_id)


def secrets_manager_factory(db: StoreInterface) -> SecretsManager:
    secrets_manager_type = CONFIG.secrets_manager_type
    match secrets_manager_type:
        case SecretsManagerType.SQL:
            return SQLSecretsManager(db)

        case _:
            raise ValueError(f"Unknown secrets manager type: {secrets_manager_type}")



from typing import Annotated
from httpx import Client
from pydantic import BaseModel

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.secret.secret import register_secret


@register_secret(secret_type="SentinelOne")
class SentinelOneSecret(BaseModel):
    base_url: str
    api_key: str


def get_sentinel_one_client(secret: SentinelOneSecret) -> Client:
    return Client(
        base_url=f"{secret.base_url}/web/api/v2.1",
        headers={
            "Authorization": f"ApiToken {secret.api_key}",
            "Accept": "application/json",
            "Content-Type": "application/json",
        },
    )


@action(
    display_name="List Alerts",
    display_namespace="SentinelOne",
    description="List alerts from SentinelOne",
    secrets_placeholders=["SENTINEL_ONE_SECRET"],
)
def list_sentinel_one_alerts(
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of cases to list.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # https://github.com/fragtastic/sentinelone-api-python/blob/67f8005576a6613f925edca93e22c8da7d6c3010/sentineloneapi/client.py#L98

    secret = ctx.get().secrets.get("SENTINEL_ONE_SECRET")
    secret = SentinelOneSecret.model_validate(secret)

    params = {
        "createdAt__gte": start_time,
        "createdAt__lte": end_time,
    }

    with get_sentinel_one_client(secret) as client:
        alerts = []

        while len(alerts) < limit:
            response = client.get("/cloud-detection/alerts", params=params)
            response.raise_for_status()
            result = response.json()

            alerts.extend(result.get("data", []))
            next_cursor = result.get("pagination", {}).get("nextCursor")

            if not next_cursor:
                break

            params["cursor"] = next_cursor

        return alerts[:limit]



import ast
import inspect
from typing import TypeVar, Callable, Any, TYPE_CHECKING
import astor
from collections import deque, defaultdict
from copy import deepcopy
from enum import Enum
import tempfile

from admyral.action_registry import ActionRegistry
from admyral.typings import JsonValue
from admyral.models import (
    NodeBase,
    ActionNode,
    UnaryOperator,
    BinaryOperator,
    Condition,
    ConstantConditionExpression,
    UnaryConditionExpression,
    BinaryConditionExpression,
    AndConditionExpression,
    OrConditionExpression,
    IfNode,
    WorkflowDAG,
    WorkflowStart,
    WorkflowWebhookTrigger,
    WorkflowScheduleTrigger,
    WorkflowDefaultArgument,
)
from admyral.compiler.condition_compiler import condition_to_str

if TYPE_CHECKING:
    from admyral.workflow import Workflow


F = TypeVar("F", bound=Callable[..., Any])


START_NODE_ID: str = "start"


class EdgeType(str, Enum):
    DEFAULT = "default"
    TRUE = "true"
    FALSE = "false"


# TODO: Introduce WorkflowCompileError
class WorkflowCompiler:
    def __init__(self) -> None:
        self.variable_to_nodes_mapping: dict[str, list[tuple[str, EdgeType]]] = (
            defaultdict(list)
        )
        self.nodes: dict[str, NodeBase] = {}

    def compile(self, workflow: "Workflow") -> WorkflowDAG:
        if inspect.iscoroutinefunction(workflow.func):
            raise RuntimeError("Async workflows are not yet supported.")

        tree = ast.parse(inspect.getsource(workflow.func))
        function_node = tree.body[0]

        start_node = ActionNode.build_start_node()
        self.nodes[START_NODE_ID] = start_node

        self._check_function_header_for_payload_parameter(workflow.func)
        self.variable_to_nodes_mapping["payload"] = [(START_NODE_ID, EdgeType.DEFAULT)]
        # Compile the workflow
        self._compile_body(function_node.body)

        return self._into_dag(workflow)

    def compile_from_module(self, module_code: str, workflow_name: str) -> WorkflowDAG:
        with tempfile.NamedTemporaryFile() as tmp:
            tmp.write(module_code.encode())
            tmp.seek(0)
            compiled_code = compile(module_code, tmp.name, "exec")
            module_namespace = {}
            exec(compiled_code, module_namespace)
            return module_namespace[workflow_name].compile()

    def compile_if_condition_str(self, condition: str) -> Condition:
        tree = ast.parse(condition)
        if len(tree.body) != 1:
            raise ValueError("Invalid condition string.")
        cond, _ = self._compile_condition_expression(tree.body[0].value)
        return cond

    def _handle_triggers(
        self, workflow: "Workflow"
    ) -> list[WorkflowWebhookTrigger | WorkflowScheduleTrigger]:
        from admyral.workflow import Webhook, Schedule

        contains_webhook = False
        triggers = []
        for trigger in workflow.triggers:
            default_args = [
                WorkflowDefaultArgument(name=k, value=v)
                for k, v in trigger.default_args.items()
            ]

            if isinstance(trigger, Webhook):
                if contains_webhook:
                    raise ValueError("A workflow can only have one webhook trigger.")
                contains_webhook = True
                triggers.append(WorkflowWebhookTrigger(default_args=default_args))
                continue

            if isinstance(trigger, Schedule):
                if trigger.cron:
                    triggers.append(
                        WorkflowScheduleTrigger(
                            cron=trigger.cron, default_args=default_args
                        )
                    )
                    continue

                if trigger.interval_seconds:
                    triggers.append(
                        WorkflowScheduleTrigger(
                            interval_seconds=trigger.interval_seconds,
                            default_args=default_args,
                        )
                    )
                    continue

                if trigger.interval_minutes:
                    triggers.append(
                        WorkflowScheduleTrigger(
                            interval_minutes=trigger.interval_minutes,
                            default_args=default_args,
                        )
                    )
                    continue

                if trigger.interval_hours:
                    triggers.append(
                        WorkflowScheduleTrigger(
                            interval_hours=trigger.interval_hours,
                            default_args=default_args,
                        )
                    )
                    continue

                if trigger.interval_days:
                    triggers.append(
                        WorkflowScheduleTrigger(
                            interval_days=trigger.interval_days,
                            default_args=default_args,
                        )
                    )
                    continue

        return triggers

    def _into_dag(self, workflow: "Workflow") -> WorkflowDAG:
        return WorkflowDAG(
            name=workflow.name,
            description=workflow.description,
            controls=self._filter_empty_controls(workflow.controls),
            start=WorkflowStart(
                triggers=self._handle_triggers(workflow),
            ),
            dag=self.nodes,
        )

    def _filter_empty_controls(self, controls: list[str] | None) -> list[str] | None:
        if not controls:
            return None
        return [control for control in controls if control]

    def _compute_unique_node_id(self, base: str) -> str:
        if base not in self.nodes:
            return base
        for idx in range(1, 1_000):
            id = f"{base}_{idx}"
            if id not in self.nodes:
                return id
        raise RuntimeError(f"Could not find a unique id for node {base}")

    def _check_function_header_for_payload_parameter(self, function: F) -> None:
        workflow_args = inspect.signature(function).parameters
        if len(workflow_args) != 1 or "payload" not in workflow_args:
            raise RuntimeError(
                f'Failed to compile workflow "{function.__name__}" because the workflow function must have exactly one parameter called "payload" and no default value: payload: dict[str, JsonValue]'
            )

        arg_name = workflow_args["payload"].name
        annotation = str(workflow_args["payload"].annotation)
        default = workflow_args["payload"].default

        if (
            arg_name != "payload"
            or annotation != "dict[str, JsonValue]"
            or default != inspect.Parameter.empty
        ):
            raise RuntimeError(
                f'Failed to compile workflow "{function.__name__}" because the workflow function must have exactly one parameter called "payload" and no default value: payload: dict[str, JsonValue]'
            )

    def _connect_node_to_dependencies(
        self,
        node_id: str,
        dependencies: set[str],
        variable_to_nodes_mapping: dict[str, list[tuple[str, EdgeType]]],
    ) -> None:
        for dependency_variable in dependencies:
            for parent_node_id, edge_type in variable_to_nodes_mapping[
                dependency_variable
            ]:
                match edge_type:
                    case EdgeType.DEFAULT:
                        self.nodes[parent_node_id].add_edge(node_id)
                    case EdgeType.TRUE:
                        self.nodes[parent_node_id].add_true_edge(node_id)
                    case EdgeType.FALSE:
                        self.nodes[parent_node_id].add_false_edge(node_id)

    def _compile_body(self, body: list[ast.stmt]) -> None:
        for statement in body:
            if isinstance(statement, ast.Assign):
                self._compile_assign(statement)

            elif isinstance(statement, ast.If):
                if_node_id, emitted_variables, leaf_nodes, if_block_dependencies = (
                    self._compile_if(statement, self.variable_to_nodes_mapping)
                )

                # Connect the if-condition node to the dependencies of the if-block
                if len(if_block_dependencies) > 0:
                    self._connect_node_to_dependencies(
                        if_node_id,
                        if_block_dependencies,
                        self.variable_to_nodes_mapping,
                    )
                else:
                    # No dependencies, add edge from start node
                    self.nodes[0].add_edge(if_node_id)

                # If an action depends on a variable emitted by the if-statement, then we
                # need to conect the action to all the leaf nodes to guarantee for
                # correct execution order because the entire if-condition block must be executed
                # before the action.
                for emitted_variable in emitted_variables:
                    self.variable_to_nodes_mapping[emitted_variable] = leaf_nodes

            elif isinstance(statement, ast.Expr):
                self._compile_expr(statement)

            else:
                raise RuntimeError(
                    f"Unsupported statement in workflow function: {astor.to_source(statement)}"
                )

    def _is_descendant(self, node: str, target: str) -> bool:
        """
        Does a path from node to target exist in the DAG?
        """
        stack = [node]
        while len(stack) > 0:
            current = stack.pop()
            if current == target:
                return True
            if isinstance(self.nodes[current], IfNode):
                stack.extend(self.nodes[current].true_children)
                stack.extend(self.nodes[current].false_children)
            else:
                stack.extend(self.nodes[current].children)
        return False

    def _remove_transitive_dependencies(
        self,
        dependencies: set[str],
        variable_to_nodes_mapping: dict[str, list[tuple[str, EdgeType]]],
    ) -> set[str]:
        """
        Suppose we have an action A which has dependencies on B and C while B has a dependency on C.

                C --
                |   |
                v   |
                B   |
                |   |
                v   |
                A <-

        Due to the transitive nature of dependencies, we can remove C from the dependencies of A
        because due to the dependency on B, A already has an implicit dependency on C.

                C --> B --> A
        """
        removal_set = set()
        for var1 in dependencies:
            # Why is [0] sufficient?
            # Case 1: if var1 is emitted by an action, then len(variable_to_nodes_mapping[var1]) == 1
            # Case 2: if var1 is emitted by an if-condition, then len(variable_to_nodes_mapping[var1]) > 1
            #         because we remember the leaf nodes of the if-condition block as representative for var1.
            #         If an action depends on var1, then it must depend on all the leaf nodes
            #         and, therefore, connects to all of the leaf nodes. Hence, doing one hop starting from any leaf node
            #         will lead to the same action nodes. Consequently, we can just use [0] to get the first representative.
            ancestor_node_id, _ = variable_to_nodes_mapping[var1][0]
            for var2 in dependencies:
                descendant_node_id, _ = variable_to_nodes_mapping[var2][0]
                if ancestor_node_id == descendant_node_id:
                    continue

                if self._is_descendant(ancestor_node_id, descendant_node_id):
                    # if var1 is a descendant of var2, then we can remove var1 because
                    # dependencies are transitive
                    removal_set.add(var1)

        return dependencies - removal_set

    def _compile_call(
        self,
        call: ast.Call,
        variable_to_nodes_mapping: dict[str, list[tuple[str, EdgeType]]],
    ) -> tuple[str, set[str]]:
        function_name = call.func.id
        function_args = call.args
        function_kwargs = call.keywords

        if len(function_args) > 0:
            raise RuntimeError(
                f"Failed to compile {function_name} because positional arguments are not supported. Please use named arguments instead."
            )

        # Check if function_name is a registered action
        if not ActionRegistry.is_registered(function_name):
            raise RuntimeError(
                f'Function "{function_name}" is not an action. Please register the function as an action if you want to use it in the workflow.'
            )

        action_secrets_mapping = {}
        dependencies = set()
        args = {}

        processed_kwargs = set()
        for kwarg in function_kwargs:
            # check for duplicate named arguments
            if kwarg.arg in processed_kwargs:
                raise RuntimeError(
                    f"Failed to compile {function_name} due to duplicate named argument: {kwarg.arg}"
                )
            processed_kwargs.add(kwarg.arg)

            if kwarg.arg == "run_after":
                # add the dependencies enforced by the run_after argument
                run_after = kwarg.value.elts
                for dependency in run_after:
                    if dependency.id not in variable_to_nodes_mapping:
                        raise RuntimeError(
                            f"Failed to compile {function_name} due to unknown dependency: {dependency.id}"
                        )
                    dependencies.add(dependency.id)
                continue

            if kwarg.arg == "secrets":
                # extract the secrets mapping from the secrets argument
                action_secrets_mapping = {
                    secret_placeholder.value: secret_id.value
                    for secret_placeholder, secret_id in zip(
                        kwarg.value.keys, kwarg.value.values
                    )
                }
                continue

            # collect dependencies and construct input args
            arg_name = kwarg.arg
            arg_value = kwarg.value
            new_dependencies, args[arg_name] = self._compile_value(arg_value)
            dependencies |= new_dependencies

        # check that for each placeholder a secret is specified
        registered_action = ActionRegistry.get(function_name)
        if len(action_secrets_mapping) != len(registered_action.secrets_placeholders):
            raise RuntimeError(
                f"Failed to compile {function_name} because provided secrets do not match secrets placeholders."
            )
        for secret_placeholder in registered_action.secrets_placeholders:
            if secret_placeholder not in action_secrets_mapping:
                raise RuntimeError(
                    f"Failed to compile {function_name} because for placeholder {secret_placeholder} no secret is defined."
                )

        # Remove transitive dependencies to avoid redundant edges in the DAG.
        dependencies = self._remove_transitive_dependencies(
            dependencies, variable_to_nodes_mapping
        )

        node_id = self._compute_unique_node_id(function_name)
        node = ActionNode(
            id=node_id,
            type=function_name,
            args=args,
            secrets_mapping=action_secrets_mapping,
        )
        self.nodes[node_id] = node

        return node_id, dependencies

    def _compile_value(self, expr: ast.expr) -> tuple[set[str], JsonValue]:
        if isinstance(expr, ast.Name):
            return {expr.id}, f"{{{{ {expr.id} }}}}"

        if isinstance(expr, ast.Subscript):
            variable_id, access_path = self._compile_subscript(expr)
            return {variable_id}, access_path

        if isinstance(expr, ast.Constant):
            return set(), expr.value

        if isinstance(expr, ast.List):
            return self._compile_list(expr)

        if isinstance(expr, ast.Dict):
            return self._compile_dict(expr)

        if isinstance(expr, ast.JoinedStr):
            return self._compile_joined_str(expr)

        raise RuntimeError(
            f"Failed to compile {astor.to_source(expr)}. Unsupported argument type."
        )

    def _compile_list(self, list_expr: ast.List) -> tuple[set[str], JsonValue]:
        compiled_json_array = []
        dependencies = set()

        for value in list_expr.elts:
            deps, compiled_value = self._compile_value(value)
            dependencies |= deps
            compiled_json_array.append(compiled_value)

        return dependencies, compiled_json_array

    def _compile_dict(self, dict_expr: ast.Dict) -> tuple[set[str], JsonValue]:
        compiled_json_object = {}
        dependencies = set()

        for key, value in zip(dict_expr.keys, dict_expr.values):
            if not isinstance(key, (ast.Constant, ast.Name, ast.Subscript)):
                raise RuntimeError(
                    f"Failed to compile {astor.to_source(dict_expr)}. Dictionary keys must be constants or string variables."
                )

            deps_key, compiled_key = self._compile_value(key)
            deps_value, compiled_value = self._compile_value(value)

            dependencies |= deps_key
            dependencies |= deps_value

            compiled_json_object[compiled_key] = compiled_value

        return dependencies, compiled_json_object

    def _compile_joined_str(self, joined_str: ast.JoinedStr) -> tuple[set[str], str]:
        dependencies = set()
        output = ""

        for value in joined_str.values:
            if isinstance(value, ast.Constant):
                output += value.value
                continue

            if isinstance(value, ast.FormattedValue):
                if value.conversion != -1:
                    raise RuntimeError(
                        f"Failed to compile {astor.to_source(joined_str)}Conversions are not supported."
                    )

                if isinstance(value.value, ast.Name):
                    dependencies.add(value.value.id)
                    output += f"{{{{ {value.value.id} }}}}"
                    continue

                if isinstance(value.value, ast.Subscript):
                    variable_id, access_path = self._compile_subscript(value.value)
                    dependencies.add(variable_id)
                    output += access_path
                    continue

                raise RuntimeError(
                    f"Failed to compile {astor.to_source(joined_str)}Only variables and subscripts are supported in f-strings."
                )

            raise RuntimeError(
                f"Failed to compile {astor.to_source(joined_str)}Unsupported value in f-string."
            )

        return dependencies, output

    def _get_assign_target(self, assign: ast.Assign) -> str:
        # => we only support a single target
        if len(assign.targets) != 1:
            raise RuntimeError(
                f"Failed to compile {astor.to_source(assign)}Only single target assignments are supported."
            )
        target = assign.targets[0]
        if not isinstance(target, ast.Name):
            raise RuntimeError(
                f"Failed to compile {astor.to_source(assign)}Only single target assignments are supported."
            )
        return target.id

    def _compile_assign(self, assign: ast.Assign) -> None:
        # Expect: NAME = CALL

        # Handle lhs: NAME
        target_variable_id = self._get_assign_target(assign)

        # we might have an await statement. unwrap it
        if isinstance(assign.value, ast.Await):
            func_call = assign.value.value
        else:
            func_call = assign.value

        if not isinstance(func_call, ast.Call):
            raise RuntimeError(
                f"Failed to compile {astor.to_source(assign)}Assignments to variables are only supported for function calls."
            )

        # Compile Action Call

        action_node_id, dependencies = self._compile_call(
            func_call, self.variable_to_nodes_mapping
        )

        # We must also consider the target variable as a dependency.
        # Consider the following workflow:
        # a = act1()
        # a = act2()
        # b = act3(a)
        #
        # If we would not add the target variable as a dependency, then act1 and act2 would be executed in parallel.
        # This would lead to a race condition because act1 could be slower than act2 and overwrite its result.
        # Hence, act3 would use the result of act1 instead of act2.
        if target_variable_id in self.variable_to_nodes_mapping:
            dependencies.add(target_variable_id)

        # Dependencies must be executed before the current action. Hence, add edges from dependencies to the current node.
        if len(dependencies) > 0:
            self._connect_node_to_dependencies(
                action_node_id, dependencies, self.variable_to_nodes_mapping
            )
        else:
            # No dependencies, add edge from start node
            self.nodes[START_NODE_ID].add_edge(action_node_id)

        self.nodes[action_node_id].result_name = target_variable_id

        # Remember the node for the target variable
        # If future actions use target_variable_id then they depend on this node
        self.variable_to_nodes_mapping[target_variable_id] = [
            (action_node_id, EdgeType.DEFAULT)
        ]

    def _compile_subscript(self, subscript: ast.Subscript) -> tuple[str, str]:
        """
        Compiles a subscript into a reference:

        Example:

            x["a"][1]["c"]

            is compiled to:

            {{ x["a"][1]["c"] }}
        """
        # extract id from Subscript
        current = subscript
        # iterate until we hit the base variable
        while not isinstance(current, ast.Name):
            current = current.value
        variable_id = current.id

        # build the access path
        access_path = ast.unparse(subscript)

        return variable_id, f"{{{{ {access_path} }}}}"

    def _compile_condition_expression(
        self,
        condition_expr: ast.expr,
    ) -> tuple[Condition, set[str]]:
        # AND / OR
        if isinstance(condition_expr, ast.BoolOp):
            deps = set()
            compiled_values = []
            for value in condition_expr.values:
                expr, new_deps = self._compile_condition_expression(value)
                deps |= new_deps
                compiled_values.append(expr)
            if isinstance(condition_expr.op, ast.And):
                return AndConditionExpression(and_expr=compiled_values), deps
            if isinstance(condition_expr.op, ast.Or):
                return OrConditionExpression(or_expr=compiled_values), deps
            raise RuntimeError(
                f"Unsupported boolean operator: {astor.to_source(condition_expr)}"
            )

        # COMPARE: LHS BINARY_OP RHS / LHS is not None / LHS is None
        if isinstance(condition_expr, ast.Compare):
            if len(condition_expr.ops) != 1 or len(condition_expr.comparators) != 1:
                raise RuntimeError(
                    f"Failed to compile {astor.to_source(condition_expr)}Only single comparison operations are supported."
                )

            lhs, deps_lhs = self._compile_condition_expression(condition_expr.left)

            # LHS is not None / LHS is None
            if isinstance(condition_expr.ops[0], (ast.Is, ast.IsNot)):
                op = (
                    UnaryOperator.IS_NONE
                    if isinstance(condition_expr.ops[0], ast.Is)
                    else UnaryOperator.IS_NOT_NONE
                )
                return UnaryConditionExpression(op=op, expr=lhs), deps_lhs

            rhs, deps_rhs = self._compile_condition_expression(
                condition_expr.comparators[0]
            )

            op_node = condition_expr.ops[0]
            if isinstance(op_node, ast.Eq):
                op = BinaryOperator.EQUALS
            elif isinstance(op_node, ast.NotEq):
                op = BinaryOperator.NOT_EQUALS
            elif isinstance(op_node, ast.Gt):
                op = BinaryOperator.GREATER_THAN
            elif isinstance(op_node, ast.Lt):
                op = BinaryOperator.LESS_THAN
            elif isinstance(op_node, ast.GtE):
                op = BinaryOperator.GREATER_THAN_OR_EQUAL
            elif isinstance(op_node, ast.LtE):
                op = BinaryOperator.LESS_THAN_OR_EQUAL
            elif isinstance(op_node, ast.In):
                op = BinaryOperator.IN
            elif isinstance(op_node, ast.NotIn):
                op = BinaryOperator.NOT_IN
            else:
                raise RuntimeError(
                    f"Unsupported comparison operator: {astor.to_source(op_node)}"
                )

            return BinaryConditionExpression(
                lhs=lhs, op=op, rhs=rhs
            ), deps_lhs | deps_rhs

        # NOT
        if isinstance(condition_expr, ast.UnaryOp):
            if isinstance(condition_expr.op, ast.Not):
                op = UnaryOperator.NOT
            else:
                raise RuntimeError(
                    f"Unsupported unary operator: {astor.to_source(condition_expr.op)}"
                )
            expr, deps = self._compile_condition_expression(condition_expr.operand)
            return UnaryConditionExpression(op=op, expr=expr), deps

        # VARIABLE
        if isinstance(condition_expr, ast.Name):
            return ConstantConditionExpression(
                value=f"{{{{ {condition_expr.id} }}}}"
            ), {condition_expr.id}

        # JSON CONSTANT
        if isinstance(condition_expr, ast.Constant):
            return ConstantConditionExpression(value=condition_expr.value), set()

        # SUBSCRIPT/REFERENCE
        if isinstance(condition_expr, ast.Subscript):
            variable_id, access_path = self._compile_subscript(condition_expr)
            return ConstantConditionExpression(value=access_path), {variable_id}

        # LIST / DICT
        if isinstance(condition_expr, (ast.List, ast.Dict)):
            deps, value = self._compile_value(condition_expr)
            return ConstantConditionExpression(value=value), deps

        raise RuntimeError(
            f"Unsupported condition expression: {astor.to_source(condition_expr)}"
        )

    def _collect_leaves_and_emitted_variables(
        self, if_node_id: str
    ) -> tuple[list[tuple[str, EdgeType]], set[str]]:
        # Perform BFS to collect all the leaf nodes and emitted variables
        # NOTE: leaf nodes means that we collect the last action node of each possible branch within the if-condition block

        leaf_nodes = set()
        emitted_variables = set()

        queue = deque([if_node_id])

        while len(queue) > 0:
            current_node_id = queue.popleft()
            current_node = self.nodes[current_node_id]

            if isinstance(current_node, ActionNode):
                if current_node.result_name is not None:
                    emitted_variables.add(current_node.result_name)

                if len(current_node.children) == 0:
                    # leaf node
                    leaf_nodes.add((current_node_id, EdgeType.DEFAULT))
                else:
                    queue.extend(current_node.children)

            else:
                # IfNode
                if len(current_node.false_children) > 0:
                    queue.extend(current_node.false_children)
                else:
                    # if the branch is empty, we still add (if_node_if, EdgeType.TRUE) to the leaf nodes
                    # because if an action has a dependency on the if-condition, then it must always be
                    # executed after the if-condition block.
                    #
                    # Consider the following example:
                    # a = act1()
                    # if a > 0:
                    #     a = act2()
                    # b = act3(a)
                    #
                    # If we would not add (if_node_id, EdgeType.FALSE) to the leaf nodes, then graph would look like:
                    #
                    #           <START>
                    #              |
                    #           a = act1() -
                    #              |       |
                    #           if a > 0   |
                    #             T|       |
                    #           a = act2() |
                    #              \      /
                    #            b = act3(a)
                    #
                    # This would mean that we would need to keep a longer history of the variables, so that we can
                    # connect act3(a) to a = act1()
                    # But if we add (if_node_id, EdgeType.FALSE) to the leaf nodes, then the graph would look like:
                    #
                    #           <START>
                    #              |
                    #           a = act1()
                    #              |
                    #           if a > 0
                    #           T/     \F
                    #    a = act2()    |
                    #              \   |
                    #            b = act3(a)
                    #
                    # This way, we can connect act3(a) to if a > 0, which is much simpler since we just remember
                    # the leaves from the if-condition block as a representative for a.
                    leaf_nodes.add((current_node_id, EdgeType.FALSE))

                if len(current_node.true_children) > 0:
                    queue.extend(current_node.true_children)
                else:
                    # Same as for the false branch
                    leaf_nodes.add((current_node_id, EdgeType.TRUE))

        return list(leaf_nodes), emitted_variables

    def _compile_if_branch(
        self,
        statements: list[ast.stmt],
        variable_to_nodes_mapping: dict[str, list[tuple[str, EdgeType]]],
        add_edge_to_if_node: Callable[..., None],
    ) -> set[str]:
        if_block_dependencies = set()

        if_branch_emitted_variables = set()
        local_variable_to_nodes_mapping = deepcopy(variable_to_nodes_mapping)

        for statement in statements:
            if isinstance(statement, (ast.Assign, ast.Expr)):
                # Expected: NAME = CALL or CALL

                if isinstance(statement.value, ast.Constant):
                    # Ignore constants. Example: multi-line comments
                    continue

                # we might have an await statement. unwrap it
                if isinstance(statement.value, ast.Await):
                    # TODO: enable async workflows later
                    # func_call = statement.value.value
                    raise RuntimeError(
                        f"Failed to compile {astor.to_source(statement)}Async functions are not yet supported."
                    )
                else:
                    func_call = statement.value

                if not isinstance(func_call, ast.Call):
                    raise RuntimeError(
                        f"Failed to compile {astor.to_source(func_call)}Unsupported expression."
                    )

                # Compile Action Call
                action_node_id, dependencies = self._compile_call(
                    func_call, local_variable_to_nodes_mapping
                )

                # Remember the dependencies from before the if-condition
                # => required for the edges to the if-condition node
                if_block_dependencies |= dependencies - if_branch_emitted_variables

                # For inside the branch, we only need to consider the dependencies that were
                # emitted within the branch
                dependencies = dependencies & if_branch_emitted_variables

                if len(dependencies) > 0:
                    self._connect_node_to_dependencies(
                        action_node_id, dependencies, local_variable_to_nodes_mapping
                    )
                else:
                    # No new dependencies from inside the branch, add edge from if-condition node
                    add_edge_to_if_node(action_node_id)

                # Handle assignment target
                # only required for NAME = CALL
                if isinstance(statement, ast.Assign):
                    target_variable_id = self._get_assign_target(statement)
                    self.nodes[action_node_id].result_name = target_variable_id
                    local_variable_to_nodes_mapping[target_variable_id] = [
                        (action_node_id, EdgeType.DEFAULT)
                    ]
                    if_branch_emitted_variables.add(
                        self.nodes[action_node_id].result_name
                    )

            elif isinstance(statement, ast.If):
                # Nested If-Condition

                (
                    nested_if_cond_node_id,
                    nested_if_emitted_variables,
                    nested_if_leaf_nodes,
                    nested_if_block_deps,
                ) = self._compile_if(statement, local_variable_to_nodes_mapping)

                # Connect the nested if-condition node to the dependencies of the nested if-condition block
                dependencies = if_branch_emitted_variables & nested_if_block_deps
                if len(dependencies) > 0:
                    self._connect_node_to_dependencies(
                        nested_if_cond_node_id,
                        dependencies,
                        local_variable_to_nodes_mapping,
                    )
                else:
                    # No new dependencies from inside the branch, add direct edge from if-condition node to nested if-condition node
                    add_edge_to_if_node(nested_if_cond_node_id)

                # update local_variable_to_nodes_mapping
                # => the emitted variables are represented by the leaves of the nested if-condition block
                for emitted_variable in nested_if_emitted_variables:
                    local_variable_to_nodes_mapping[emitted_variable] = (
                        nested_if_leaf_nodes
                    )

                # push the emitted variables to the if_branch_emitted_variables because
                # now the if-branch also emits the variables from the nested if-condition block
                if_branch_emitted_variables |= nested_if_emitted_variables

            else:
                raise RuntimeError(
                    f"Unsupported statement in workflow function: {astor.to_source(statement)}"
                )

        return if_block_dependencies

    def _compile_if(
        self,
        if_ast: ast.If,
        variable_to_nodes_mapping: dict[str, list[tuple[str, EdgeType]]],
    ) -> tuple[int, set[str], list[tuple[str, EdgeType]], set[str]]:
        prev_variables_to_nodes_mapping = deepcopy(variable_to_nodes_mapping)

        # compile condition + collect dependencies from the condition
        if_condition, if_block_dependencies = self._compile_condition_expression(
            if_ast.test
        )
        if_condition_str = condition_to_str(if_ast.test)

        if_node_id = self._compute_unique_node_id("if")
        if_node = IfNode(
            id=if_node_id, condition=if_condition, condition_str=if_condition_str
        )
        self.nodes[if_node_id] = if_node

        # Compile true branch
        if_block_dependencies |= self._compile_if_branch(
            if_ast.body, variable_to_nodes_mapping, if_node.add_true_edge
        )
        # Compile false branch
        if_block_dependencies |= self._compile_if_branch(
            if_ast.orelse, variable_to_nodes_mapping, if_node.add_false_edge
        )

        # remove transitive dependencies from if_block_dependencies

        # We need to remove transitive dependencies from the statements before the if-block to avoid
        # connectng the if-block to redundant edges in the DAG.
        #
        # Consider the following example:
        # a = act1()
        # b = act2(a)
        # if a > 0 and b < 0:
        #    ...
        #
        # Then, a and b are both dependencies for the if-block. But since b depends on a, we can
        # remove the edge from a to the if-condition node because the if-condition node already
        # implicilty depends on a due to the dependency on b.
        #
        # NOTE: we need to use the initial variable_to_nodes_mapping because nested if-conditions
        # might have overwritten the variable_to_nodes_mapping.
        if_block_dependencies = self._remove_transitive_dependencies(
            if_block_dependencies, prev_variables_to_nodes_mapping
        )

        # Collect leaf nodes and emitted variables from if-condition block
        leaf_nodes, emitted_variables = self._collect_leaves_and_emitted_variables(
            if_node_id
        )

        return if_node_id, emitted_variables, leaf_nodes, if_block_dependencies

    def _compile_expr(self, expr: ast.Expr) -> None:
        # Expect: CALL

        if isinstance(expr.value, ast.Constant):
            # Ignore constants. Example: multi-line comments
            return

        # we might have an await statement. unwrap it
        if isinstance(expr.value, ast.Await):
            # TODO: enable async workflows later
            # func_call = expr.value.value
            raise RuntimeError(
                f"Failed to compile {astor.to_source(expr)}Async functions are not yet supported."
            )
        else:
            func_call = expr.value

        if not isinstance(func_call, ast.Call):
            raise RuntimeError(
                f"Failed to compile {astor.to_source(func_call)}Unsupported expression."
            )

        # Compile Action Call
        action_node_id, dependencies = self._compile_call(
            func_call, self.variable_to_nodes_mapping
        )
        # Dependencies must be executed before the current action. Hence, add edges from dependencies to the current node.
        if len(dependencies) > 0:
            self._connect_node_to_dependencies(
                action_node_id, dependencies, self.variable_to_nodes_mapping
            )
        else:
            # No dependencies, add edge from start node
            self.nodes[START_NODE_ID].add_edge(action_node_id)



from msgraph import GraphServiceClient
from azure.identity import ClientSecretCredential
from enum import Enum
from msgraph.generated.security.alerts_v2.alerts_v2_request_builder import (
    Alerts_v2RequestBuilder,
)
from msgraph.generated.models.security.alert import Alert
from kiota_serialization_json.json_serialization_writer import JsonSerializationWriter
import json
from httpx import Client
from pydantic import BaseModel

from admyral.typings import JsonValue
from admyral.utils.future_executor import execute_future
from admyral.utils.collections import is_not_empty
from admyral.secret.secret import register_secret


@register_secret(secret_type="Azure")
class AzureSecret(BaseModel):
    tenant_id: str
    client_id: str
    client_secret: str


def get_ms_graph_client(
    secret: AzureSecret,
) -> GraphServiceClient:
    """
    Create an MS Graph client using the client credentials provider (OAuth 2.0)

    Sources:
        - https://learn.microsoft.com/en-us/graph/sdks/choose-authentication-providers?tabs=python#client-credentials-provider
        - https://learn.microsoft.com/en-us/entra/identity-platform/v2-oauth2-client-creds-grant-flow
    """
    return GraphServiceClient(
        ClientSecretCredential(
            tenant_id=secret.tenant_id,
            client_id=secret.client_id,
            client_secret=secret.client_secret,
        ),
        ["https://graph.microsoft.com/.default"],
    )


# Note: also exists as auto-generated enum in msgraph.generated.models.security.service_source.py
class MsSecurityGraphAlertServiceSource(str, Enum):
    MS_DEFENDER_FOR_ENDPOINT = "microsoftDefenderForEndpoint"
    MS_DEFENDER_FOR_IDENTITY = "microsoftDefenderForIdentity"
    MS_DEFENDER_FOR_CLOUD_APPS = "microsoftDefenderForCloudApps"
    MS_DEFENDER_FOR_OFFICE365 = "microsoftDefenderForOffice365"
    MS_365_DEFENDER = "microsoft365Defender"
    AZURE_AD_IDENTITY_PROTECTION = "azureAdIdentityProtection"
    MS_APP_GOVERNANCE = "microsoftAppGovernance"
    DATA_LOSS_PREVENTION = "dataLossPrevention"
    MS_DEFENDER_FOR_CLOUD = "microsoftDefenderForCloud"
    MS_SENTINEL = "microsoftSentinel"


def _alert_to_json(alert: Alert) -> JsonValue:
    writer = JsonSerializationWriter()
    alert.serialize(writer)
    return json.loads(writer.get_serialized_content().decode("utf-8"))


def ms_graph_list_alerts_v2(
    secret: AzureSecret,
    start_time: str | None = None,
    end_time: str | None = None,
    service_source: MsSecurityGraphAlertServiceSource | None = None,
    limit: int = 100,
) -> list[dict[str, JsonValue]]:
    """
    Fetch alerts from the Microsoft Graph Security API using the List Alerts v2 endpoint.

    Sources:
        - https://learn.microsoft.com/en-us/graph/api/security-list-alerts_v2?view=graph-rest-1.0&tabs=python
        - https://learn.microsoft.com/en-us/graph/auth-v2-service?tabs=http
        - https://github.com/microsoftgraph/msgraph-sdk-python/blob/main/msgraph/generated/security/alerts_v2/alerts_v2_request_builder.py
        - https://learn.microsoft.com/en-us/entra/identity-platform/v2-oauth2-client-creds-grant-flow
    """
    client = get_ms_graph_client(secret)

    filter_params = []
    if service_source:
        filter_params.append(f"serviceSource eq '{service_source.value}'")
    if start_time:
        filter_params.append(f"createdDateTime ge {start_time}")
    if end_time:
        filter_params.append(f"createdDateTime le {end_time}")

    query_params = Alerts_v2RequestBuilder.Alerts_v2RequestBuilderGetQueryParameters(
        filter=" and ".join(filter_params) if is_not_empty(filter_params) else None,
    )
    request_config = (
        Alerts_v2RequestBuilder.Alerts_v2RequestBuilderGetRequestConfiguration(
            query_parameters=query_params
        )
    )

    alert_collection = execute_future(
        client.security.alerts_v2.get(request_configuration=request_config)
    )

    alerts = [_alert_to_json(alert) for alert in alert_collection.value]

    # handle pagination
    while len(alerts) < limit and alert_collection and alert_collection.odata_next_link:
        alert_collection = execute_future(
            client.security.alerts_v2.with_url(alert_collection.odata_next_link).get(
                request_configuration=request_config
            )
        )
        alerts.extend([_alert_to_json(alert) for alert in alert_collection.value])

    return alerts[:limit]


class MSGraphClient:
    def __init__(self, secret: AzureSecret):
        self.secret = secret
        self.token = None
        self.base_url = "https://graph.microsoft.com/v1.0"
        self.client = Client()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.client.close()

    def _get_token(self) -> str:
        token_url = f"https://login.microsoftonline.com/{self.secret.tenant_id}/oauth2/v2.0/token"
        data = {
            "client_id": self.secret.client_id,
            "scope": "https://graph.microsoft.com/.default",
            "client_secret": self.secret.client_secret,
            "grant_type": "client_credentials",
        }
        response = self.client.post(token_url, data=data)
        response.raise_for_status()
        self.token = response.json()["access_token"]
        return self.token

    def _make_request(self, method: str, endpoint: str, **kwargs) -> dict:
        if not self.token:
            self._get_token()

        headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
        }
        url = f"{self.base_url}{endpoint}"

        response = self.client.request(method, url, headers=headers, **kwargs)

        if response.status_code == 401:  # Token might have expired
            self._get_token()
            headers["Authorization"] = f"Bearer {self.token}"
            response = self.client.request(method, url, headers=headers, **kwargs)

        response.raise_for_status()
        return response.json()

    def get(self, url: str) -> dict:
        return self._make_request("GET", url)


def ms_graph_list_managed_devices(secret: AzureSecret) -> list[dict[str, JsonValue]]:
    """
    Fetch managed devices from the Microsoft Graph Security API using the List Managed Devices endpoint.

    Sources:
        - https://learn.microsoft.com/en-us/graph/api/intune-devices-manageddevice-list?view=graph-rest-1.0
        - https://github.com/microsoftgraph/msgraph-sdk-python/blob/6ec0cca83fdcbde8a6cca6b8407b2ffc59343b9e/msgraph/generated/devices/item/device_item_request_builder.py
    """

    with MSGraphClient(secret) as client:
        managed_devices = []
        next_link = "/deviceManagement/managedDevices"

        while next_link:
            response = client.get(next_link)
            managed_devices.extend(response["value"])

            next_link = response.get("@odata.nextLink")
            if next_link:
                next_link = next_link.split(client.base_url)[1]

        return managed_devices



from admyral.db.schemas.pip_lockfile_cache_schemas import PipLockfileCacheSchema
from admyral.db.schemas.python_action_schemas import PythonActionSchema
from admyral.db.schemas.workflow_schemas import WorkflowSchema
from admyral.db.schemas.workflow_run_schemas import (
    WorkflowRunSchema,
    WorkflowRunStepsSchema,
)
from admyral.db.schemas.workflow_webhook_schemas import WorkflowWebhookSchema
from admyral.db.schemas.workflow_schedule_schemas import WorkflowScheduleSchema
from admyral.db.schemas.secrets_schemas import SecretsSchema
from admyral.db.schemas.auth_schemas import (
    UserSchema,
    AccountSchema,
    SessionSchema,
    VerificationTokenSchema,
    AuthenticatorSchema,
)
from admyral.db.schemas.api_key_schemas import ApiKeySchema

__all__ = [
    "PipLockfileCacheSchema",
    "PythonActionSchema",
    "WorkflowSchema",
    "WorkflowRunSchema",
    "WorkflowRunStepsSchema",
    "WorkflowWebhookSchema",
    "WorkflowScheduleSchema",
    "SecretsSchema",
    "UserSchema",
    "AccountSchema",
    "SessionSchema",
    "VerificationTokenSchema",
    "AuthenticatorSchema",
    "ApiKeySchema",
]



from typing import Annotated

from admyral.action import action, ArgumentMetadata
from admyral.context import ctx
from admyral.typings import JsonValue
from admyral.actions.integrations.shared.ms_graph import (
    ms_graph_list_alerts_v2,
    MsSecurityGraphAlertServiceSource,
    AzureSecret,
)


@action(
    display_name="List Alerts",
    display_namespace="Microsoft Defender for Cloud",
    description="List alerts from Microsoft Defender for Cloud",
    secrets_placeholders=["AZURE_SECRET"],
)
def list_ms_defender_for_cloud_alerts(
    start_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="Start Time",
            description="The start time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    end_time: Annotated[
        str | None,
        ArgumentMetadata(
            display_name="End Time",
            description="The end time for the cases to list. Must be in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).",
        ),
    ] = None,
    limit: Annotated[
        int,
        ArgumentMetadata(
            display_name="Limit",
            description="The maximum number of cases to list.",
        ),
    ] = 1000,
) -> list[dict[str, JsonValue]]:
    # https://learn.microsoft.com/en-us/rest/api/defenderforcloud/alerts/list?view=rest-defenderforcloud-2022-01-01&tabs=HTTP
    secret = ctx.get().secrets.get("AZURE_SECRET")
    secret = AzureSecret.model_validate(secret)

    return ms_graph_list_alerts_v2(
        secret=secret,
        start_time=start_time,
        end_time=end_time,
        limit=limit,
        service_source=MsSecurityGraphAlertServiceSource.MS_DEFENDER_FOR_CLOUD,
    )



from typing import Any, AsyncGenerator
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.exc import IntegrityError
from sqlmodel import select, delete, insert, update
from datetime import datetime
from sqlmodel.ext.asyncio.session import AsyncSession
from sqlalchemy.orm import sessionmaker
from contextlib import asynccontextmanager
from uuid import uuid4
import json

from admyral.models import (
    User,
    PipLockfile,
    PythonAction,
    Workflow,
    WorkflowRun,
    WorkflowWebhook,
    EncryptedSecret,
    WorkflowMetadata,
    ActionMetadata,
    SecretMetadata,
    WorkflowRunMetadata,
    WorkflowRunStepMetadata,
    WorkflowRunStep,
    ApiKey,
)
from admyral.models.workflow_schedule import WorkflowSchedule
from admyral.db.store_interface import StoreInterface
from admyral.db.schemas import (
    PythonActionSchema,
    PipLockfileCacheSchema,
    WorkflowSchema,
    WorkflowRunSchema,
    WorkflowRunStepsSchema,
    WorkflowWebhookSchema,
    WorkflowScheduleSchema,
    SecretsSchema,
    UserSchema,
    ApiKeySchema,
)
from admyral.db.alembic.database_manager import DatabaseManager
from admyral.config.config import GlobalConfig, CONFIG
from admyral.logger import get_logger
from admyral.utils.time import utc_now
from admyral.utils.crypto import generate_hs256
from admyral.typings import JsonValue


logger = get_logger(__name__)


class AdmyralDatabaseSession:
    def __init__(self, session: AsyncSession, execution_options: dict = {}) -> None:
        self.session = session
        self.execution_options = execution_options

    @classmethod
    async def from_session(
        cls,
        session: AsyncSession,
        execution_options: dict = {},
    ) -> "AdmyralDatabaseSession":
        return cls(session, execution_options)

    async def exec(self, statement: Any) -> Any:
        return await self.session.exec(
            statement, execution_options=self.execution_options
        )

    async def commit(self) -> None:
        await self.session.commit()


class AdmyralStore(StoreInterface):
    def __init__(self, config: GlobalConfig) -> None:
        self.config = config

        self.engine = create_async_engine(
            self.config.database_url, echo=True, future=True, pool_pre_ping=True
        )
        self.async_session_maker = sessionmaker(
            self.engine, class_=AsyncSession, expire_on_commit=False
        )

        self.execution_options = {}

        self.performed_setup = False

    # TODO: pass down config
    @classmethod
    async def create_store(cls, skip_setup: bool = False) -> "AdmyralStore":
        store = cls(CONFIG)
        if not skip_setup:
            await store.setup()
        store.performed_setup = True

        return store

    ########################################################
    # Setup & Chore
    ########################################################

    async def setup(self):
        logger.info("Setting up Admyral store.")

        database_manager = DatabaseManager(self.engine, self.config)

        does_db_exist = await database_manager.database_exists()
        if not does_db_exist:
            logger.info("Creating database...")
            await database_manager.create_database()
            logger.info("Database created.")

        logger.info("Running migrations...")
        await database_manager.migrate_database()
        logger.info("Migrations complete.")

        logger.info("Setting up Admyral store complete.")

    ########################################################
    # Helpers
    ########################################################

    @asynccontextmanager
    async def _get_async_session(self) -> AsyncGenerator[AdmyralDatabaseSession, None]:
        assert self.performed_setup, "Store has not been set up yet."
        async with self.async_session_maker() as session:
            db = await AdmyralDatabaseSession.from_session(
                session,
                self.execution_options,
            )
            yield db

    ########################################################
    # User Management
    ########################################################

    async def get_user(self, user_id: str) -> User | None:
        async with self._get_async_session() as db:
            result = await db.exec(select(UserSchema).where(UserSchema.id == user_id))
            return result.one_or_none()

    async def clean_up_workflow_data_of(self, user_id: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(WorkflowSchema).where(WorkflowSchema.user_id == user_id)
            )
            await db.commit()

    async def delete_user(self, user_id: str) -> None:
        async with self._get_async_session() as db:
            # Note: all associated data is deleted via cascade
            await db.exec(delete(UserSchema).where(UserSchema.id == user_id))
            await db.commit()

    ########################################################
    # API Key Management
    ########################################################

    async def store_api_key(self, user_id: str, name: str, key: str) -> ApiKey:
        async with self._get_async_session() as db:
            api_key_id = str(uuid4())
            await db.exec(
                insert(ApiKeySchema).values(
                    id=api_key_id,
                    user_id=user_id,
                    name=name,
                    key=key,
                )
            )
            await db.commit()

            result = await db.exec(
                select(ApiKeySchema, UserSchema)
                .join(UserSchema)
                .where(ApiKeySchema.user_id == user_id)
                .where(ApiKeySchema.id == api_key_id)
                .where(UserSchema.id == user_id)
            )
            row = result.one()

            return ApiKey(
                id=row[0].id,
                name=row[0].name,
                created_at=row[0].created_at,
                user_email=row[1].email,
            )

    async def list_api_keys(self, user_id: str) -> list[ApiKey]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(ApiKeySchema, UserSchema)
                .join(UserSchema)
                .where(ApiKeySchema.user_id == user_id)
            )
            return list(
                map(
                    lambda row: ApiKey.model_validate(
                        {
                            "id": row[0].id,
                            "name": row[0].name,
                            "created_at": row[0].created_at,
                            "user_email": row[1].email,
                        }
                    ),
                    result.all(),
                )
            )

    async def search_api_key_owner(self, key: str) -> str | None:
        async with self._get_async_session() as db:
            result = await db.exec(select(ApiKeySchema).where(ApiKeySchema.key == key))
            api_key = result.one_or_none()
            return api_key.user_id if api_key else None

    async def delete_api_key(self, user_id: str, key_id: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(ApiKeySchema)
                .where(ApiKeySchema.user_id == user_id)
                .where(ApiKeySchema.id == key_id)
            )
            await db.commit()

    ########################################################
    # Python Action
    ########################################################

    async def list_actions(self, user_id: str) -> list[ActionMetadata]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(PythonActionSchema).where(PythonActionSchema.user_id == user_id)
            )
            return [action.to_metadata() for action in result.all()]

    async def _get_action(
        self, db: AdmyralDatabaseSession, user_id: str, action_type: str
    ) -> PythonActionSchema | None:
        result = await db.exec(
            select(PythonActionSchema)
            .where(PythonActionSchema.action_type == action_type)
            .where(PythonActionSchema.user_id == user_id),
        )
        return result.one_or_none()

    async def get_action(self, user_id: str, action_type: str) -> PythonAction | None:
        async with self._get_async_session() as db:
            action = await self._get_action(db, user_id, action_type)
            return action.to_model() if action is not None else None

    async def store_action(self, user_id: str, action: PythonAction) -> None:
        async with self._get_async_session() as db:
            # if the action type already exists, we update the code and requirements.
            python_action_entry = await self._get_action(
                db, user_id, action.action_type
            )

            secrets_placeholders = (
                ";".join(action.secrets_placeholders)
                if action.secrets_placeholders
                else None
            )
            requirements = (
                ";".join(action.requirements) if action.requirements else None
            )

            if python_action_entry:
                await db.exec(
                    update(PythonActionSchema)
                    .where(PythonActionSchema.action_type == action.action_type)
                    .where(PythonActionSchema.user_id == user_id)
                    .values(
                        import_statements=action.import_statements,
                        code=action.code,
                        display_name=action.display_name,
                        display_namespace=action.display_namespace,
                        description=action.description,
                        secrets_placeholders=secrets_placeholders,
                        requirements=requirements,
                        arguments=[arg.model_dump() for arg in action.arguments],
                        updated_at=utc_now(),
                    )
                )
            else:
                await db.exec(
                    insert(PythonActionSchema).values(
                        user_id=user_id,
                        action_type=action.action_type,
                        import_statements=action.import_statements,
                        code=action.code,
                        display_name=action.display_name,
                        display_namespace=action.display_namespace,
                        description=action.description,
                        secrets_placeholders=secrets_placeholders,
                        requirements=requirements,
                        arguments=[arg.model_dump() for arg in action.arguments],
                    )
                )

            await db.commit()

    async def delete_action(self, user_id: str, action_type: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(PythonActionSchema)
                .where(PythonActionSchema.action_type == action_type)
                .where(PythonActionSchema.user_id == user_id)
            )
            await db.commit()

    ########################################################
    # Pip Lockfile Cache
    ########################################################

    async def _get_cached_pip_lockfile(
        self, db: AdmyralDatabaseSession, hash: str
    ) -> PipLockfileCacheSchema | None:
        result = await db.exec(
            select(PipLockfileCacheSchema).where(PipLockfileCacheSchema.hash == hash)
        )
        return result.one_or_none()

    async def get_cached_pip_lockfile(self, hash: str) -> PipLockfile | None:
        async with self._get_async_session() as db:
            pip_lockfile = await self._get_cached_pip_lockfile(db, hash)
            return pip_lockfile.to_model() if pip_lockfile else None

    async def cache_pip_lockfile(
        self,
        pip_lockfile: PipLockfile,
    ) -> None:
        try:
            async with self._get_async_session() as db:
                # check whether the lockfile is already cached. if yes, then we update
                # its expiration time. otherwise, we create a new cache entry.
                cached_pip_lockfile = await self._get_cached_pip_lockfile(
                    db, pip_lockfile.hash
                )

                if cached_pip_lockfile:
                    await db.exec(
                        update(PipLockfileCacheSchema)
                        .where(PipLockfileCacheSchema.hash == pip_lockfile.hash)
                        .values(
                            expiration_time=datetime.fromtimestamp(
                                pip_lockfile.expiration_time
                            ),
                            updated_at=utc_now(),
                        )
                    )
                else:
                    await db.exec(
                        insert(PipLockfileCacheSchema).values(
                            hash=pip_lockfile.hash,
                            lockfile=pip_lockfile.lockfile,
                            expiration_time=datetime.fromtimestamp(
                                pip_lockfile.expiration_time
                            ),
                        )
                    )

                await db.commit()
        except IntegrityError:
            # In case of a race condition, we might receive a unique constraint violation.
            # We can safely ignore this error.
            logger.warning(
                f"Non-critical race condition detected during caching lockfile with hash {pip_lockfile.hash}. Ignoring."
            )

    async def delete_expired_cached_pip_lockfile(self) -> None:
        async with self._get_async_session() as db:
            current_time = utc_now()
            await db.exec(
                delete(PipLockfileCacheSchema).where(
                    PipLockfileCacheSchema.expiration_time < current_time
                )
            )
            await db.commit()

    ########################################################
    # Workflows
    ########################################################

    async def list_workflows(self, user_id: str) -> list[WorkflowMetadata]:
        async with self._get_async_session() as db:
            # TODO: loading all workflows just for the metadata
            # is not very efficient
            result = await db.exec(
                select(WorkflowSchema).where(WorkflowSchema.user_id == user_id)
            )
            return [workflow.to_metadata() for workflow in result.all()]

    async def _get_workflow_by_name(
        self, db: AdmyralDatabaseSession, user_id: str, workflow_name: str
    ) -> WorkflowSchema | None:
        result = await db.exec(
            select(WorkflowSchema)
            .where(WorkflowSchema.workflow_name == workflow_name)
            .where(WorkflowSchema.user_id == user_id)
        )
        return result.one_or_none()

    async def get_workflow_by_name(
        self, user_id: str, workflow_name: str
    ) -> Workflow | None:
        async with self._get_async_session() as db:
            wf = await self._get_workflow_by_name(db, user_id, workflow_name)
            return wf.to_model() if wf else None

    async def _get_workflow_by_id(
        self, db: AdmyralDatabaseSession, user_id: str, workflow_id: str
    ) -> WorkflowSchema | None:
        result = await db.exec(
            select(WorkflowSchema)
            .where(WorkflowSchema.workflow_id == workflow_id)
            .where(WorkflowSchema.user_id == user_id)
        )
        return result.one_or_none()

    async def get_workflow_by_id(
        self, user_id: str, workflow_id: str
    ) -> Workflow | None:
        async with self._get_async_session() as db:
            wf = await self._get_workflow_by_id(db, user_id, workflow_id)
            return wf.to_model() if wf else None

    async def store_workflow(self, user_id: str, workflow: Workflow) -> None:
        async with self._get_async_session() as db:
            # if the workflow already exists, we update the workflow dag.
            stored_workflow = await self._get_workflow_by_id(
                db, user_id, workflow.workflow_id
            )

            if stored_workflow:
                await db.exec(
                    update(WorkflowSchema)
                    .where(WorkflowSchema.user_id == user_id)
                    .where(WorkflowSchema.workflow_id == workflow.workflow_id)
                    .values(
                        workflow_name=workflow.workflow_name,  # consider workflow name changes
                        workflow_dag=workflow.workflow_dag.model_dump(),
                        is_active=workflow.is_active,
                    )
                )
            else:
                await db.exec(
                    insert(WorkflowSchema).values(
                        workflow_id=workflow.workflow_id,
                        user_id=user_id,
                        workflow_name=workflow.workflow_name,
                        workflow_dag=workflow.workflow_dag.model_dump(),
                        is_active=workflow.is_active,
                    )
                )

            await db.commit()

    async def set_workflow_active_state(
        self, user_id: str, workflow_id: str, is_active: bool
    ) -> None:
        async with self._get_async_session() as db:
            stored_workflow = await self._get_workflow_by_id(db, user_id, workflow_id)
            if not stored_workflow:
                raise ValueError(f"Workflow {workflow_id} not found.")

            await db.exec(
                update(WorkflowSchema)
                .where(WorkflowSchema.user_id == user_id)
                .where(WorkflowSchema.workflow_id == workflow_id)
                .values(is_active=is_active)
            )
            await db.commit()

    async def remove_workflow(self, user_id: str, workflow_id: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(WorkflowSchema)
                .where(WorkflowSchema.user_id == user_id)
                .where(WorkflowSchema.workflow_id == workflow_id)
            )
            await db.commit()

    async def get_workflow_for_webhook(
        self, workflow_id: str
    ) -> tuple[str, Workflow] | None:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(WorkflowSchema).where(WorkflowSchema.workflow_id == workflow_id)
            )
            workflow = result.one_or_none()
            return (workflow.user_id, workflow.to_model()) if workflow else None

    ########################################################
    # Workflow Webhooks
    ########################################################

    async def store_workflow_webhook(
        self, user_id: str, workflow_id: str
    ) -> WorkflowWebhook:
        webhook_id = str(uuid4())
        webhook_secret = generate_hs256(webhook_id)

        async with self._get_async_session() as db:
            # verify that the user_id owns the workflow_id
            workflow = await self._get_workflow_by_id(db, user_id, workflow_id)
            if not workflow:
                raise ValueError("Failed to store webhook. Workflow not found.")

            await db.exec(
                insert(WorkflowWebhookSchema).values(
                    webhook_id=webhook_id,
                    workflow_id=workflow_id,
                    webhook_secret=webhook_secret,
                )
            )
            await db.commit()

        return WorkflowWebhook(
            webhook_id=webhook_id,
            workflow_id=workflow_id,
            webhook_secret=webhook_secret,
        )

    async def get_webhook_for_workflow(
        self, user_id: str, workflow_id: str
    ) -> WorkflowWebhook | None:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(WorkflowWebhookSchema)
                .join(WorkflowSchema)
                .where(WorkflowSchema.user_id == user_id)
                .where(WorkflowSchema.workflow_id == workflow_id)
                .where(WorkflowWebhookSchema.workflow_id == workflow_id)
            )
            webhooks = result.all()

            if len(webhooks) == 0:
                return None
            if len(webhooks) > 1:
                # This should not happen!
                raise ValueError(f"Multiple webhooks found for workflow {workflow_id}")

            return webhooks[0].to_model()

    async def _get_webhook(
        self, db: AdmyralDatabaseSession, webhook_id: str
    ) -> WorkflowWebhook | None:
        result = await db.exec(
            select(WorkflowWebhookSchema).where(
                WorkflowWebhookSchema.webhook_id == webhook_id
            )
        )
        return result.one_or_none()

    async def get_webhook(self, webhook_id: str) -> WorkflowWebhook | None:
        async with self._get_async_session() as db:
            webhook = await self._get_webhook(db, webhook_id)
            return webhook.to_model() if webhook else None

    async def delete_webhook(self, user_id: str, webhook_id: str) -> None:
        async with self._get_async_session() as db:
            # verify that user_id owns the webhook
            workflow = await db.exec(
                select(WorkflowSchema)
                .join(WorkflowWebhookSchema)
                .where(WorkflowSchema.user_id == user_id)
                .where(WorkflowWebhookSchema.webhook_id == webhook_id)
            )
            if not workflow:
                raise ValueError("Failed to delete webhook. Workflow not found.")

            await db.exec(
                delete(WorkflowWebhookSchema).where(
                    WorkflowWebhookSchema.webhook_id == webhook_id
                )
            )
            await db.commit()

    ########################################################
    # Workflow Schedules
    ########################################################

    async def store_schedule(self, user_id: str, schedule: WorkflowSchedule) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                insert(WorkflowScheduleSchema).values(
                    schedule_id=schedule.schedule_id,
                    workflow_id=schedule.workflow_id,
                    user_id=user_id,
                    cron=schedule.cron,
                    interval_seconds=schedule.interval_seconds,
                    interval_minutes=schedule.interval_minutes,
                    interval_hours=schedule.interval_hours,
                    interval_days=schedule.interval_days,
                    default_args=schedule.default_args,
                )
            )
            await db.commit()

    async def list_schedules_for_workflow(
        self, user_id: str, workflow_id: str
    ) -> list[WorkflowSchedule]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(WorkflowScheduleSchema)
                .where(WorkflowScheduleSchema.user_id == user_id)
                .where(WorkflowScheduleSchema.workflow_id == workflow_id)
            )
            schedules = result.all()
            return [schedule.to_model() for schedule in schedules]

    async def delete_schedule(self, user_id: str, schedule_id: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(WorkflowScheduleSchema)
                .where(WorkflowScheduleSchema.user_id == user_id)
                .where(WorkflowScheduleSchema.schedule_id == schedule_id)
            )
            await db.commit()

    ########################################################
    # Workflow Runs - User Facing
    ########################################################

    async def list_workflow_runs(
        self, user_id: str, workflow_id: str, limit: int = 100
    ) -> list[WorkflowRunMetadata]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(WorkflowRunSchema)
                .join(WorkflowSchema)
                .where(WorkflowSchema.user_id == user_id)
                .where(WorkflowRunSchema.workflow_id == workflow_id)
                .order_by(WorkflowRunSchema.created_at.desc())
                .limit(limit)
            )
            return [workflow_run.to_metadata() for workflow_run in result.all()]

    async def _get_workflow_run(
        self, db: AdmyralDatabaseSession, user_id: str, workflow_id: str, run_id: str
    ) -> WorkflowRunSchema | None:
        result = await db.exec(
            select(WorkflowRunSchema)
            .join(WorkflowSchema)
            .where(WorkflowSchema.user_id == user_id)
            .where(WorkflowRunSchema.workflow_id == workflow_id)
            .where(WorkflowRunSchema.run_id == run_id)
        )
        return result.one_or_none()

    async def get_workflow_run(
        self, user_id: str, workflow_id: str, run_id: str
    ) -> WorkflowRun | None:
        async with self._get_async_session() as db:
            workflow_run = await self._get_workflow_run(
                db, user_id, workflow_id, run_id
            )
            return workflow_run.to_model() if workflow_run else None

    async def list_workflow_run_steps(
        self, user_id: str, workflow_id: str, run_id: str
    ) -> list[WorkflowRunStepMetadata]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(
                    WorkflowRunStepsSchema.step_id,
                    WorkflowRunStepsSchema.action_type,
                    WorkflowRunStepsSchema.error,
                )
                .join(WorkflowRunSchema)
                .join(WorkflowSchema)
                .where(WorkflowRunSchema.workflow_id == workflow_id)
                .where(WorkflowSchema.user_id == user_id)
                .where(WorkflowSchema.workflow_id == workflow_id)
                .where(WorkflowRunStepsSchema.run_id == run_id)
                .order_by(WorkflowRunStepsSchema.created_at)
            )
            return [
                WorkflowRunStepMetadata.model_validate(
                    {"step_id": row[0], "action_type": row[1], "error": row[2]}
                )
                for row in result.all()
            ]

    async def _get_workflow_run_step(
        self,
        db: AdmyralDatabaseSession,
        user_id: str,
        workflow_id: str,
        run_id: str,
        step_id: str,
    ) -> WorkflowRunStepsSchema | None:
        result = await db.exec(
            select(WorkflowRunStepsSchema)
            .join(WorkflowRunSchema)
            .join(WorkflowSchema)
            .where(WorkflowRunSchema.run_id == run_id)
            .where(WorkflowRunSchema.workflow_id == workflow_id)
            .where(WorkflowSchema.workflow_id == workflow_id)
            .where(WorkflowSchema.user_id == user_id)
            .where(WorkflowRunStepsSchema.step_id == step_id)
        )
        return result.one_or_none()

    async def get_workflow_run_step(
        self, user_id: str, workflow_id: str, run_id: str, step_id: str
    ) -> WorkflowRunStep | None:
        async with self._get_async_session() as db:
            workflow_run_step = await self._get_workflow_run_step(
                db, user_id, workflow_id, run_id, step_id
            )
            return workflow_run_step.to_model() if workflow_run_step else None

    ########################################################
    # Workflow Runs - State Updates during execution
    ########################################################

    async def init_workflow_run(
        self,
        run_id: str,
        step_id: str,
        workflow_id: str,
        source_name: str,
        payload: dict[str, JsonValue],
    ) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                insert(WorkflowRunSchema).values(
                    run_id=run_id,
                    workflow_id=workflow_id,
                    source_name=source_name,
                )
            )

            await db.exec(
                insert(WorkflowRunStepsSchema).values(
                    step_id=step_id, run_id=run_id, action_type="start", result=payload
                )
            )

            await db.commit()

    async def mark_workflow_run_as_completed(
        self, run_id: str, completed_at: datetime
    ) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                update(WorkflowRunSchema)
                .where(WorkflowRunSchema.run_id == run_id)
                .values(
                    completed_at=completed_at,
                )
            )
            await db.commit()

    async def _get_workflow_step_without_user_id(
        self,
        db: AdmyralDatabaseSession,
        run_id: str,
        step_id: str,
    ) -> WorkflowRunStepsSchema | None:
        result = await db.exec(
            select(WorkflowRunStepsSchema)
            .where(WorkflowRunStepsSchema.step_id == step_id)
            .where(WorkflowRunStepsSchema.run_id == run_id)
        )
        return result.one_or_none()

    async def append_logs(
        self,
        step_id: str,
        run_id: str,
        action_type: str,
        prev_step_id: str,
        lines: list[str],
    ) -> None:
        async with self._get_async_session() as db:
            workflow_run_step = await self._get_workflow_step_without_user_id(
                db, run_id, step_id
            )

            if not workflow_run_step:
                await db.exec(
                    insert(WorkflowRunStepsSchema).values(
                        step_id=step_id,
                        run_id=run_id,
                        action_type=action_type,
                        prev_step_id=prev_step_id,
                        logs="\n".join(lines),
                    )
                )
            else:
                await db.exec(
                    update(WorkflowRunStepsSchema)
                    .where(WorkflowRunStepsSchema.step_id == step_id)
                    .values(logs=workflow_run_step.logs + "\n" + "\n".join(lines))
                )
            await db.commit()

    async def store_action_result(
        self,
        step_id: str,
        run_id: str,
        action_type: str,
        prev_step_id: str,
        result: JsonValue,
        input_args: dict[str, JsonValue],
    ) -> None:
        async with self._get_async_session() as db:
            workflow_run_step = await self._get_workflow_step_without_user_id(
                db, run_id, step_id
            )

            if not workflow_run_step:
                await db.exec(
                    insert(WorkflowRunStepsSchema).values(
                        step_id=step_id,
                        run_id=run_id,
                        action_type=action_type,
                        prev_step_id=prev_step_id,
                        result=result,
                        input_args=input_args,
                    )
                )
            else:
                await db.exec(
                    update(WorkflowRunStepsSchema)
                    .where(WorkflowRunStepsSchema.step_id == step_id)
                    .values(
                        result=result,
                        input_args=input_args,
                    )
                )
            await db.commit()

    async def store_workflow_run_error(
        self,
        step_id: str,
        run_id: str,
        action_type: str,
        prev_step_id: str,
        error: str,
        input_args: dict[str, JsonValue],
    ) -> None:
        async with self._get_async_session() as db:
            workflow_run_step = await self._get_workflow_step_without_user_id(
                db, run_id, step_id
            )

            if not workflow_run_step:
                await db.exec(
                    insert(WorkflowRunStepsSchema).values(
                        step_id=step_id,
                        run_id=run_id,
                        action_type=action_type,
                        prev_step_id=prev_step_id,
                        input_args=input_args,
                        error=error,
                    )
                )
            else:
                await db.exec(
                    update(WorkflowRunStepsSchema)
                    .where(WorkflowRunStepsSchema.step_id == step_id)
                    .values(
                        error=error,
                        input_args=input_args,
                    )
                )

            await db.exec(
                update(WorkflowRunSchema)
                .where(WorkflowRunSchema.run_id == run_id)
                .values(failed_at=utc_now())
            )

            await db.commit()

    ########################################################
    # Secrets
    ########################################################

    async def list_secrets(
        self,
        user_id: str,
    ) -> list[SecretMetadata]:
        async with self._get_async_session() as db:
            result = await db.exec(
                select(SecretsSchema, UserSchema)
                .join(UserSchema)
                .where(SecretsSchema.user_id == user_id)
                .where(UserSchema.id == user_id)
            )
            return list(
                map(
                    lambda row: SecretMetadata.model_validate(
                        {
                            "secret_id": row[0].secret_id,
                            "secret_schema": json.loads(row[0].schema_json_serialized),
                            "email": row[1].email,
                            "created_at": row[0].created_at,
                            "updated_at": row[0].updated_at,
                            "secret_type": row[0].secret_type,
                        }
                    ),
                    result.all(),
                ),
            )

    async def _get_secret_and_owner(
        self, db: AdmyralDatabaseSession, user_id: str, secret_id: str
    ) -> tuple | None:
        result = await db.exec(
            select(SecretsSchema, UserSchema)
            .join(UserSchema)
            .where(SecretsSchema.user_id == user_id)
            .where(SecretsSchema.secret_id == secret_id)
            .where(UserSchema.id == user_id)
        )
        return result.one_or_none()

    async def get_secret(self, user_id: str, secret_id: str) -> EncryptedSecret | None:
        async with self._get_async_session() as db:
            secret_and_user = await self._get_secret_and_owner(db, user_id, secret_id)
            return secret_and_user[0].to_model() if secret_and_user else None

    async def store_secret(
        self,
        user_id: str,
        secret_id: str,
        encrypted_secret: str | None,
        schema: list[str],
        secret_type: str | None = None,
    ) -> SecretMetadata:
        async with self._get_async_session() as db:
            secret_and_user = await self._get_secret_and_owner(db, user_id, secret_id)

            if secret_and_user:
                secret_result = await db.exec(
                    update(SecretsSchema)
                    .where(SecretsSchema.user_id == user_id)
                    .where(SecretsSchema.secret_id == secret_id)
                    .values(
                        encrypted_secret=encrypted_secret,
                        schema_json_serialized=json.dumps(schema),
                        updated_at=utc_now(),
                    )
                    .returning(SecretsSchema.created_at, SecretsSchema.updated_at)
                )
                secret_created_at_updated_at = secret_result.one()

                user_email = secret_and_user[1].email
            else:
                secret_result = await db.exec(
                    insert(SecretsSchema)
                    .values(
                        user_id=user_id,
                        secret_id=secret_id,
                        encrypted_secret=encrypted_secret,
                        schema_json_serialized=json.dumps(schema),
                        secret_type=secret_type,
                    )
                    .returning(SecretsSchema.created_at, SecretsSchema.updated_at)
                )
                secret_created_at_updated_at = secret_result.one()

                user_result = await db.exec(
                    select(UserSchema).where(UserSchema.id == user_id).limit(1)
                )
                user_email = user_result.one().email

            await db.commit()

            return SecretMetadata(
                secret_id=secret_id,
                secret_schema=schema,
                email=user_email,
                created_at=secret_created_at_updated_at[0],
                updated_at=secret_created_at_updated_at[1],
                secret_type=secret_type,
            )

    async def delete_secret(self, user_id: str, secret_id: str) -> None:
        async with self._get_async_session() as db:
            await db.exec(
                delete(SecretsSchema)
                .where(SecretsSchema.user_id == user_id)
                .where(SecretsSchema.secret_id == secret_id)
            )
            await db.commit()

    async def compare_and_swap_secret(
        self,
        user_id: str,
        secret_id: str,
        expected_encrypted_secret: str | None,
        encrypted_secret: str | None,
        secret_schema: list[str],
        secret_type: str | None = None,
    ) -> SecretMetadata:
        async with self._get_async_session() as db:
            secret_and_owner = await self._get_secret_and_owner(db, user_id, secret_id)
            if not secret_and_owner:
                raise ValueError(
                    f"Failed to update secret because the secret {secret_id} was not found."
                )
            if secret_and_owner[0].encrypted_secret != expected_encrypted_secret:
                raise ValueError(
                    "Failed to update secret because the secret was modified simultaneously."
                )

            secret_result = await db.exec(
                update(SecretsSchema)
                .where(SecretsSchema.user_id == user_id)
                .where(SecretsSchema.secret_id == secret_id)
                .values(
                    encrypted_secret=encrypted_secret,
                    schema_json_serialized=json.dumps(secret_schema),
                    updated_at=utc_now(),
                )
                .returning(SecretsSchema.created_at, SecretsSchema.updated_at)
            )
            secret_created_at_updated_at = secret_result.one()

            await db.commit()

            return SecretMetadata(
                secret_id=secret_id,
                secret_schema=secret_schema,
                email=secret_and_owner[1].email,
                created_at=secret_created_at_updated_at[0],
                updated_at=secret_created_at_updated_at[1],
                secret_type=secret_type,
            )
